
import sys,os,re,time
from functools import wraps
from selenium.webdriver.remote.webelement import WebElement

from blues_lib.util.BluesConsole import BluesConsole

class InfoDeco():

  caller_class = 'Info'

  def __init__(self,caller_method,level=3):
    '''
    Create a special decorator for class waiter.EC 
    Parameter:
      caller_method {str} : the wrappered func's name
    '''
    self.caller_method = caller_method
    self.level = level

    # func's args index
    self.arg_cs_index = None
    self.arg_pcs_index = None

    # rewrite the index
    self.set_arg_index()
  
  def __call__(self,func):
    @wraps(func) 
    def wrapper(*args,**kwargs):
      return self.wrapper(func,*args,**kwargs)
    return wrapper

  def set_arg_index(self):
    '''
    Template method
    '''
    self.arg_key_index = None
    self.arg_cs_index = 1
    self.arg_pcs_index = 2

  def wrapper(self,func,*args,**kwargs):

    # execute the wrappered func
    outcome = func(*args,**kwargs)

    msg = self.get_finish_msg(outcome,args)

    # print success msg
    if self.level<=2:
      BluesConsole.info(msg)
    
    # must return the wrappered func's value
    return outcome

  def get_finish_msg(self,outcome,args):
    info = self.get_info(args)
    info.update({ 
      'type':type(outcome).__name__,
      'value': self.get_value(outcome)
    })

    values = (self.caller_class,self.caller_method,info)
    return '[%s.%s] Result %s' % values

  def get_value(self,outcome):
    if not outcome:
      return outcome 
    elif type(outcome) == str:
      if len(outcome)<=30:
        return outcome
      else:
        return outcome[0:30]+'...'
    else:
      return str(outcome)
   
  def get_info(self,args):
    key = self.get_key(args)
    cs = self.get_cs(args)
    pcs = self.get_pcs(args)
    info = {}
    if key:
      info['key'] = key
    if cs:
      info['cs'] = cs
    if pcs:
      info['pcs'] = pcs
    return info

  def get_key(self,args):
    index = self.arg_key_index
    if index and len(args)>index:
      return args[index]
    else:
      return None

  def get_cs(self,args):
    index = self.arg_cs_index
    return self.get_cs_title(args,index)

  def get_pcs(self,args):
    index = self.arg_pcs_index
    return self.get_cs_title(args,index)

  def get_cs_title(self,args,index):
    if index and len(args)>index:
      CS_WE = args[index]
      if type(CS_WE) == WebElement:
        return 'WebElement %s' % CS_WE.tag_name
      else:
        return CS_WE
    else:
      return None





from functools import wraps
from blues_lib.type.output.STDOut import STDOut
from blues_lib.deco.LogDeco import LogDeco

class CommandSTDOutLog(LogDeco):
  def __init__(self,title:str='node',max_chars:int=-1):
    super().__init__(title,max_chars)

  def __call__(self,func):
    @wraps(func) 
    def wrapper(instance,*args,**kwargs):
      # this is the decoed class instance
      stdout:STDOut = func(instance,*args,**kwargs)
      prefix:str = self._get_prefix(instance)
      self._log(prefix,stdout)
      return stdout

    return wrapper
  
  def _get_prefix(self,instance)->str:
    module_name:str = instance.__class__.__module__  # 模块路径（如 my_project.tasks.loop）
    task_id:str = instance._ti.task_id
    round:int = 0
    if self._title == 'loop' and hasattr(instance, "_round"):
      round = instance._round
    round_str:str = ''
    if round>0:
      round_str = f'round {round}'
    return f'{module_name} {task_id} [{self._title}] {round_str}'



import sys,os,re

from blues_lib.sele.waiter.Querier import Querier  
from selenium.webdriver import ActionChains
from selenium.webdriver.common.actions.wheel_input import ScrollOrigin

class Wheel():
  '''
  A representation of a scroll wheel input device for interacting with a web page.
  '''
  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver)
    self.__chains = ActionChains(driver)

  def scroll_to_element(self,loc_or_elem,parent_loc_or_elem=None):
    '''
    The actions class does not automatically scroll the target element into view,
    So this method will need to be used if elements are not already inside the viewport.
    The viewport will be scrolled so the bottom of the element is at the bottom of the screen.
    Parameter:
      loc_or_elem {str | WebElement} : css selector or web element
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem)
    if not web_element:
      return None

    self.__chains.scroll_to_element(web_element).perform()

  def scroll_to_viewport(self,amount_x=0,amount_y=0):
    '''
    Scroll by given amount
    This is the second most common scenario for scrolling. 
    Pass in an delta x and a delta y value for how much to scroll in the right and down directions. 
    Negative values represent left and up, respectively.
    Parameter:
      amount_x {int} : the movement amount on  x-axis
      amount_y {int} : the movement amount on  y-axis
    '''
    self.__chains.scroll_by_amount(amount_x,amount_y).perform()
    
  def scroll_element_to_center(self,loc_or_elem,parent_loc_or_elem=None,timeout=None):

    element = self.__querier.query(loc_or_elem,parent_loc_or_elem)
    if not element:
      return None

    # scroll the element's center to the center of the viewport
    # 1. 获取窗口尺寸及中心点坐标
    window_size = self.__driver.get_window_size()
    window_center_x = window_size['width'] // 2   # 窗口水平中心
    window_center_y = window_size['height'] // 2  # 窗口垂直中心
    
    # 2. 获取元素自身尺寸及中心点坐标（相对于元素左上角的偏移）
    element_rect = element.rect  # 包含width/height/x/y等信息
    element_center_x = element_rect['width'] // 2  # 元素水平中心（相对于自身左上角）
    element_center_y = element_rect['height'] // 2 # 元素垂直中心（相对于自身左上角）
    
    # 3. 计算需要滚动的距离
    # 原理：从元素左上角开始，滚动到「窗口中心 - 元素中心」的位置
    scroll_x = window_center_x - element_center_x
    scroll_y = window_center_y - element_center_y

    self.scroll_from_element_to_offset(loc_or_elem,scroll_x,scroll_y,parent_loc_or_elem)

  def scroll_from_element_to_offset(self,loc_or_elem,amount_x=0,amount_y=0,parent_loc_or_elem=None):
    '''
    Scroll from an element by a given amount
    This scenario is effectively a combination of the above two methods.
    If the element is out of the viewport, it will be scrolled to the bottom of the screen, 
      then the page will be scrolled by the provided delta x and delta y values.
    Parameter:
      amount_x {int} : the movement amount on  x-axis
      amount_y {int} : the movement amount on  y-axis
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem)
    if not web_element:
      return None
    scroll_origin = ScrollOrigin.from_element(web_element)
    self.__chains.scroll_from_origin(scroll_origin,amount_x,amount_y).perform()


  def scroll_from_element_offset_to_offset(self,loc_or_elem,offset_x,offset_y,amount_x,amount_y,parent_loc_or_elem=None):
    '''
    Scroll from an element with an offset
    This scenario is used when you need to scroll only a portion of the screen, and it is outside the viewport. Or is inside the viewport and the portion of the screen that must be scrolled is a known offset away from a specific element.
    This uses the “Scroll From” method again, and in addition to specifying the element, an offset is specified to indicate the origin point of the scroll. The offset is calculated from the center of the provided element.
    If the element is out of the viewport, it first will be scrolled to the bottom of the screen, then the origin of the scroll will be determined by adding the offset to the coordinates of the center of the element, and finally the page will be scrolled by the provided delta x and delta y values.
    Parameter:
      offset_x {int} : the offset value on  x-axis to the element's cetner point
      offset_y {int} : the offset value on  y-axis to the element's cetner point
      amount_x {int} : the movement amount on  x-axis
      amount_y {int} : the movement amount on  y-axis
    Exception:
      If the offset from the center of the element falls outside of the viewport, it will result in an exception.
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem)
    if not web_element:
      return None
    
    offset_origin = ScrollOrigin.from_element(web_element,offset_x,offset_y)
    self.__chains.scroll_from_origin(offset_origin,amount_x,amount_y).perform()

  def scroll_from_viewport_offset_to_offset(self,offset_x,offset_y,amount_x,amount_y):
    '''
    Scroll from a offset of origin (viewport) by given amount
    The final scenario is used when you need to scroll only a portion of the screen, and it is already inside the viewport.
    This uses the “Scroll From” method again, but the viewport is designated instead of an element. An offset is specified from the upper left corner of the current viewport. After the origin point is determined, the page will be scrolled by the provided delta x and delta y values.
    Parameter:
      offset_x {int} : the offset value on  x-axis to the element's cetner point
      offset_y {int} : the offset value on  y-axis to the element's cetner point
    Exception:
      If the offset from the upper left corner of the viewport falls outside of the screen, it will result in an exception.
    '''
    offset_origin = ScrollOrigin.from_viewport(offset_x,offset_y)
    self.__chains.scroll_from_origin(offset_origin,amount_x,amount_y).perform()






import pandas

class BluesCSV():
  
  @classmethod
  def read(cls,csv_file):
    df = pandas.read_csv(csv_file)
    return df

  @classmethod
  def read_head(cls,csv_file):
    '''
    Returns {str} : multi original lines at the beginning
    '''
    df = cls.read(csv_file)
    return df.head()

  @classmethod
  def read_rows(cls,csv_file):
    df = cls.read(csv_file)
    return df.to_dict(orient='records')
  
  @classmethod
  def write(cls,file_path,rows,columns=None,index=False):
    '''
    Create or cover a csv file
    @param {str} file_path : the csv file abs path
    @param {list<list> | list<dict>} rows 
    @param {list<str>} columns : the csv's header - first row
      - if the item is a dict, don't need 
    @param {bool} index : create a index column
    '''
    if isinstance(rows[0],dict):
      df = pd.DataFrame(rows)
    else:
      df = pd.DataFrame(rows, columns=columns)
    df.to_csv(file_path, index=index) 

  @classmethod
  def append(cls,file_path,rows,index=False):
    '''
    Append rows to a exists csv file
    @param {str} file_path : the csv file abs path
    @param {list<list> | list<dict>} rows 
    @param {list<str>} columns : the csv's header - first row
      - if the item is a dict, don't need 
    @param {bool} index : create a index column
    '''
    df = pd.DataFrame(rows)
    df.to_csv(file_path,mode='a',header=False,index=index) 



import sys,os,re
from typing import Any

from blues_lib.behavior.Bean import Bean

class Style(Bean):

  def _get(self)->str:
    kwargs = self._get_kwargs(['loc_or_elem','parent_loc_or_elem','timeout'])
    kwargs.update({'key':'style'})
    return self._browser.element.info.get_attr(**kwargs)


  def _set(self):
    selector:str = self._config.get('loc_or_elem')
    if entity := self._get_value_entity():
      attrs = {
        "style":self._get_inline_style(entity),
      }
      if self._to_be_presence():
        return self._browser.script.javascript.attr(selector,attrs)
    
  def _get_inline_style(self,styles: dict) -> str:
    """
    将字典形式的CSS样式转换为字符串形式
    
    参数:
        styles: 包含CSS样式的字典，键可以是驼峰式(如fontWeight)或连字符式(如font-size)
    
    返回:
        转换后的CSS样式字符串，所有键都会转换为连字符形式
    """
    style_items = []
    
    for key, value in styles.items():
      # 将驼峰式命名转换为连字符式 (如fontWeight -> font-weight)
      converted_key = []
      for char in key:
        if char.isupper():
          # 在大写字母前添加连字符，并将大写转为小写
          converted_key.append('-' + char.lower())
        else:
          converted_key.append(char)
      css_key = ''.join(converted_key)
      
      # 添加键值对到列表
      style_items.append(f"{css_key}:{value}")
    
    # 用分号连接所有样式项
    return ';'.join(style_items)


from blues_lib.sele.waiter.Querier import Querier  
from blues_lib.sele.element.Info import Info  
from blues_lib.type.file.File import File
from blues_lib.util.FileDownloader import FileDownloader    

class Image():

  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver,5) 
    self.__info = Info(driver) 

  def download_image(self,image_CS_WE,image_dir=None,parent_loc_or_elem=None,timeout=5):
    '''
    Returns:
      {dict} : format download output, like:
        {'code':200,files:[],'message':''}
    '''
    image_CS_WEs = image_CS_WE if type(image_CS_WE)==list else [image_CS_WE]
    urls = self.get_img_urls(image_CS_WEs,parent_loc_or_elem,timeout)
    if urls:
      default_dir = File.get_dir_path(['download','img']) 
      file_dir = image_dir if image_dir else default_dir
      return FileDownloader.download(urls,file_dir)
    else:
      return None

  def get_img_urls(self,image_CS_WE,parent_loc_or_elem=None,timeout=5):
    '''
    Get all img urls from multiple selectors
    '''
    if not image_CS_WE:
      return None
    
    image_CS_WEs = image_CS_WE if type(image_CS_WE)==list else [image_CS_WE]

    urls = []
    for image_CS_WE in image_CS_WEs:
      if not image_CS_WE:
        continue
      current_urls = self.__get_img_urls(image_CS_WE,parent_loc_or_elem,timeout)
      if current_urls:
        urls.extend(current_urls)
    return urls

  def __get_img_urls(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Support img or other elements
    Returns:
      {list<str>} : the url list
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return None
    urls = []
    if web_element.tag_name == 'img':
      url = self.__info.get_attr(web_element,'src')
      if url:
        urls.append(url)
    else:
      img_elements = self.__querier.query_all('img',web_element,timeout)
      if img_elements:
        for img_element in img_elements:
          url = self.__info.get_attr(img_element,'src')
          if url:
            urls.append(url)
    return urls





from blues_lib.namespace.CommandName import CommandName
from blues_lib.dao.sql.TableMutator import TableMutator
from blues_lib.type.output.STDOut import STDOut
from blues_lib.command.SQLCommand import SQLCommand

class Updater(SQLCommand):

  NAME = CommandName.SQL.UPDATER

  def _invoke(self)->STDOut:
    table:str = self._summary.get('table')
    entities:list[dict] = self._summary.get('entities')
    mutator = TableMutator(table)
    
    results:list[dict] = []
    is_failed:bool = False
    for entity in entities:
      result:STDOut = self._invoke_one(mutator,entity)
      results.append(result.to_dict())
      if result.code != 200:
        is_failed = True
    
    return STDOut(200,'ok',results) if not is_failed else STDOut(500,'failed',results)

  def _invoke_one(self,mutator:TableMutator,entity:dict)->STDOut:
    conditions = self._get_conditions(entity)
    filter_rule = self._get_filter_rule()
    filtered_entities = self._filter_entities([entity],filter_rule)
    return mutator.update(filtered_entities[0],conditions)
  
  def _get_conditions(self,entity:dict)->list[dict]|None:
    if conditions := self._summary.get('conditions'):
      return conditions

    if condition_field := self._summary.get('condition_field'):
      condition_value = entity.get(condition_field)
      return [{
        'field':condition_field,
        'comparator':'=',
        'value':condition_value
      }]
    return None
  
  def _get_filter_rule(self)->dict:
    filter_rule:dict = self._summary.copy()
    if condition_field:=self._summary.get('condition_field'):
      # exclude the condition field
      filter_rule['exc_fields'] = [condition_field]
    return filter_rule



from blues_lib.command.FlowCommand import FlowCommand
from blues_lib.namespace.CommandName import CommandName
from blues_lib.type.output.STDOut import STDOut

class Engine(FlowCommand):

  NAME = CommandName.Flow.ENGINE

  def _run(self)->STDOut: 
    dag_biz:list[dict] = self._bizdata.get('tasks')
    return self._run_flow(self._dag_def,dag_biz)
  


from blues_lib.type.output.STDOut import STDOut
from .Completion import Completion

class Chat(Completion):
  _STREAM = False
  
  def _get_content(self,response)->STDOut:
    if not response.choices or not response.choices[0].message.content:
      raise Exception(f'{self.__class__.__name__} no choices[0].message.content')

    return self._get_output(response)

  def _get_output(self,response)->STDOut:
    content:str = response.choices[0].message.content
    return STDOut(
      200,
      'success',
      content,
      response.usage,
    )




from abc import ABC,abstractmethod
from typing import final

from blues_lib.type.executor.Command import Command
from blues_lib.namespace.CrawlerName import CrawlerName
from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model
from blues_lib.command.io.IOExcept import IOExcept
from blues_lib.command.io.InputHandler import InputHandler
from blues_lib.command.io.OutputHandler import OutputHandler
from blues_lib.hook.command.CommandHook import CommandHook
from blues_lib.deco.CommandSTDOutLog import CommandSTDOutLog

class NodeCommand(Command,ABC):

  NAME = None
  
  def __init__(self,task_def:dict,bizdata:dict,ti:any,context:dict|None=None) -> None:
    f'''
    Args:
      task_def {dict}: the task's definition
        - id {str} : the task id
        - command {str} : the command name
        - meta {dict} : the task meta data
        - input {list|dict|None} : the task input definition
        - output {list|dict|None} : the task output definition
        - setup {list|dict|None} : the task setup hook
        - teardown {list|dict|None} : the task teardown hook
      bizdata {dict} : the task bizdata
      ti {any} : the task instance
    '''
    super().__init__({})

    # the input mapping can not useing the placeholder
    self._task_def:dict = task_def
    self._bizdata:dict = bizdata
    self._ti:any = ti
    self._context = context

    self._model = None
    self._meta = {}
    self._config = {}
    self._summary = {}

    self._dag_def = {}

  @property
  def task_id(self)->str:
    return self._task_def.get('task_id')
  
  def _init(self)->None:

    task_def,task_biz = InputHandler(self._task_def,self._bizdata,self._ti).handle()
    self._task_def = task_def
    self._bizdata = task_biz

    # set model or dag
    self._create_model()
    self._create_dag()

    # set fields after recalculate the model
    self._setup()

  @final
  @CommandSTDOutLog()
  def execute(self)->STDOut:

    # must validate in runtime, make sure the upstream's data has created
    self._init()

    # Airflow不能识别自定义异常，使用xcom标识状态
    hook_defs:list[dict] = self._task_def.get('before_invoked')
    options:dict = {'ti':self._ti}
    CommandHook(hook_defs,options).execute()

    # 因为有多个hook，所以不能基于某个实例判断是否block或skip
    CommandHook.block(self._ti)
    stdout:STDOut|None = CommandHook.skip(self._ti)
    if stdout:
      return stdout

    stdout = self._invoke()

    hook_defs:list[dict] = self._task_def.get('after_invoked')
    options:dict = {'ti':self._ti,'stdout':stdout}
    CommandHook(hook_defs,options).execute()

    # 因为有多个hook，所以不能基于某个实例判断是否block或skip
    CommandHook.block(self._ti)
    # output after hook deal
    OutputHandler.handle(self._ti,stdout)

    # raise except if the stdout is not matched
    IOExcept.validate_except(self._task_def,stdout)
    
    return stdout
  
  def _create_model(self):
    if meta:= self._task_def.get('meta'):
      self._meta = meta
      self._model:Model = Model(meta,self._bizdata)
      self._config:dict = self._model.config
      self._summary:dict = self._config.get(CrawlerName.Field.SUMMARY.value) or {}
      # validate the metadata
      IOExcept.validate_metadata(self._config)
      
  def _create_dag(self):
    if dag:= self._task_def.get('dag'):
      self._dag_def:dict = dag
    
  def _setup(self): 
    # template method
    pass

  @abstractmethod
  def _invoke(self)->STDOut:
    pass



from blues_lib.util.ResourceReader import ResourceReader
from blues_lib.util.DotFileReader import DotFileReader

from blues_lib.metadata import instances  # 对应 instances 目录

class MetaInstance:

  @classmethod
  def get(cls, dot_path:str) -> dict:
    extension:str = 'conf'
    slash_path:str = DotFileReader.get_slash_path(dot_path,extension)
    return ResourceReader.read_hocon(instances, slash_path)



from blues_lib.hocon.replacer.HoconReplacer import HoconReplacer

class Model:

  def __init__(self, meta: dict, bizdata: dict|None = None) -> None:
    self.meta = self._get_meta(meta)
    self.bizdata = self._get_bizdata(bizdata)
    self.config = self._get_config()
    
  def _get_meta(self,meta:dict)->dict:
    '''
    complete the macro calculation in the meta (include and function)
    - doesn't pass the vars, wouldn't replace the vars
    '''
    return HoconReplacer(meta).replace()
    
  def _get_bizdata(self,bizdata:dict|None)->dict:
    '''
    complete the macro calculation in the bizdata (include and function)
    - doesn't pass the vars, wouldn't replace the vars
    '''
    return HoconReplacer(bizdata).replace()
    
  def refresh(self):
    '''
    recalculate the config, when the meta or bizdata is updated
    '''
    self.config = self._get_config()

  def _get_config(self) -> dict:
    """
    replace the placeholder in the meta with the bizdata ("{}")
    Args:
      meta {dict}: the meta template
      bizdata {dict}: the meta variables
    Returns:
      {dict}: the config
    """
    return HoconReplacer(self.meta,self.bizdata).replace()






from importlib import resources
from blues_lib.util.FileReader import FileReader

class ResourceReader:
  """
  Universal package resource reader supporting JSON/JSON5/YAML formats.
  Accepts both package objects and string package paths as input.
  Requires Python 3.9+ and installed dependencies: json5, pyyaml.
  Encoding is fixed to utf-8.
  """

  @classmethod
  def _get_package_name(cls, package):
    if hasattr(package, "__name__"):
      return package.__name__
    elif isinstance(package, str):
      return package
    raise TypeError("package must be a package object or string package path")

  @classmethod
  def get_file_path(cls, package, resource_path):
    '''
    Get the absolute file path of a resource within a package.
    @param package: The package object or string package path.
      - use from blues_lib.util.mock import mock to get the package object.
      - use 'blues_lib.util.mock' to get the string package path.
    @param resource_path: The relative path to the resource within the package. such as 'tests/mock/command/llm-loop-urls/def.conf'
    @return: The absolute file path of the resource.
    @raises FileNotFoundError: If the resource does not exist in the package.
    '''
    package_name = cls._get_package_name(package)
    pkg_dir = resources.files(package)
    file_path = pkg_dir.joinpath(resource_path)
    if not file_path.exists():
      raise FileNotFoundError(f"Resource not found in package [{package_name}]: {resource_path}")
    return file_path

  @classmethod
  def read_hocon(cls, package, resource_path):
    file_path = cls.get_file_path(package, resource_path)
    return FileReader.read_hocon(file_path)

  @classmethod
  def read_text(cls, package, resource_path):
    file_path = cls.get_file_path(package, resource_path)
    return FileReader.read_text(file_path)

  @classmethod
  def read_json(cls, package, resource_path):
    file_path = cls.get_file_path(package, resource_path)
    return FileReader.read_json(file_path)

  @classmethod
  def read_json5(cls, package, resource_path):
    file_path = cls.get_file_path(package, resource_path)
    return FileReader.read_json5(file_path)

  @classmethod
  def read_yaml(cls, package, resource_path):
    file_path = cls.get_file_path(package, resource_path)
    return FileReader.read_yaml(file_path)

  @classmethod
  def read_csv(cls, package, resource_path):
    file_path = cls.get_file_path(package, resource_path)
    return FileReader.read_csv(file_path)

  @classmethod
  def read_binary(cls, package, resource_path):
    file_path = cls.get_file_path(package, resource_path)
    return file_path.read_bytes()
  
  


import logging
from abc import ABC

class Factory(ABC):
  
  def __init__(self):
    self._logger = logging.getLogger('airflow.task')

  def create(self, mode: str, **kwargs):
    method_name = f"create_{mode.lower()}"
    if not hasattr(self, method_name):
      return None
    method = getattr(self, method_name)
    return method(**kwargs)

  


from blues_lib.namespace.NSEnum import NSEnum
from blues_lib.namespace.EnumNS import EnumNS

class CommandName(EnumNS):
  
  class Flow(NSEnum):
    ENGINE = "command.flow.engine"
    LOOP = "command.flow.loop"

  class LLM(NSEnum):
    ENGINE = "command.llm.engine"
    LOOP = "command.llm.loop"
    
  class Crawler(NSEnum):
    ENGINE = "command.crawler.engine"
    LOOP = "command.crawler.loop"
  
  class Standard(NSEnum):
    DUMMY = "command.standard.dummy"
    CLEANER = "command.standard.cleaner"
  
  # sql 
  class SQL(NSEnum):
    QUERIER = "command.sql.querier"
    UPDATER = "command.sql.updater"
    INSERTER = "command.sql.inserter"
    DELETER = "command.sql.deleter"

  class Notifier(NSEnum):
    EMAIL = "command.notifier.email"

  class Material(NSEnum):
    SINKER = "command.material.sinker"
    DEDUPLICATOR = "command.material.deduplicator"
    VALIDATOR = "command.material.validator"
    NORMALIZER = "command.material.normalizer"
    LOCALIZER = "command.material.localizer"



from typing import TypeAlias,Any
from selenium.webdriver.remote.webelement import WebElement

LocOrElem: TypeAlias = str | WebElement | None
LocOrElems: TypeAlias = str | list[str] | list[WebElement] | None

WaitTime: TypeAlias = int | float | tuple[int | float, str] | tuple[int | float, int | float, str] | None
IntervalTime: TypeAlias = int | float | list[int | float] | list[int | float | tuple[int | float, str]] | None

LightBhvItem: TypeAlias = tuple[str,str,Any|None]
LightBhvList: TypeAlias = list[LightBhvItem] | None


from typing import Any
import json,csv,yaml,json5
from blues_lib.hocon.HoconReader import HoconReader

class FileReader:

  @classmethod
  def read(cls,file_path:str)->str:
    try:
      with open(file_path, 'r', encoding='utf-8') as file:
        return file.read()
    except Exception as e:
      print(f"read file error: {e}")
      return ''

  @classmethod
  def read_text(cls,file_path:str)->str:
    return cls.read(file_path)

  @classmethod
  def read_yaml(cls,file_path:str)->Any|None:
    try:
      with open(file_path, 'r', encoding='utf-8') as file:
        return yaml.safe_load(file)
    except Exception as e:
      print(f"read yaml error: {e}")
      return None

  @classmethod
  def read_json5(cls,file_path:str)->Any|None:
    try:
      with open(file_path, 'r', encoding='utf-8') as file:
        data = json5.load(file)
        return data
    except Exception as e:
      print(f"read json5 error: {e}")
      return None

  @classmethod
  def read_json(cls,file_path:str)->Any|None:
    try:
      with open(file_path, 'r', encoding='utf-8') as file:
        data = json.load(file)
        return data
    except Exception as e:
      print(f"read json error: {e}")
      return None

  @classmethod
  def read_csv(cls,file_path:str,headless=False)->list[list[str]]|None:
    '''
    @description : read csv file
    @param {str} file_path
    @returns {list<list>}
    '''
    try:
      with open(file_path, 'r', encoding='utf-8',errors='ignore') as file:
        lines = csv.reader(file) 
        rows = []
        i=0
        for line in lines:
          i+=1
          if i==1 and headless:
            continue
          if not line:
            continue
          rows.append(line)
        return rows
    except Exception as e:
      print(f"read csv error: {e}")
      return None

  @classmethod
  def read_hocon(cls,file_path:str)->dict:
    '''
    @description: read conf file 
    @param {str} file path such as 'c:/blues-lib-py/tests/mock/command/llm-loop-urls/def.conf'
    @returns {any|None}
    '''
    return HoconReader.read_as_dict(file_path)



import sys,os,re

from .NAPS import NAPS

from blues_lib.schema.releaser.channels.ChannelsSchemaFactory import ChannelsSchemaFactory
from blues_lib.loginer.factory.ChannelsLoginerFactory import ChannelsLoginerFactory   
from blues_lib.sele.publisher.OnceLoginPublisher import OnceLoginPublisher
from blues_lib.sele.publisher.visitor.ActivityVisitor import ActivityVisitor

# don't need spide
class ChannelsVideoNAPS():
  '''
  1. Login the publish page
  2. Upload the video and fill the from
  3. Submit
  '''

  def publish(self):
    loginer_factory = ChannelsLoginerFactory()
    loginer = loginer_factory.create_qrcode(once=True)

    factory = ChannelsSchemaFactory()
    schema = factory.create_video()
    
    # max activity count
    maxnium = 50
    # start activity index (start from 0)
    starting_index=1
    # recursive wait time seconds ,at lease 100
    recursive_interval=60*60*2
    # recursive time , max 24
    recursive_time=12

    visitor = ActivityVisitor(maxnium,starting_index,recursive_interval,recursive_time)
    publisher = OnceLoginPublisher(schema,loginer)
    publisher.accept(visitor)




import time,random
from abc import abstractmethod

from blues_lib.type.executor.Executor import Executor
from blues_lib.model.Model import Model
from blues_lib.type.output.STDOut import STDOut
from blues_lib.sele.browser.Browser import Browser
from blues_lib.hook.behavior.BehaviorHook import BehaviorHook 
from blues_lib.schema.SchemaValidator import SchemaValidator
from blues_lib.types.common import WaitTime
from blues_lib.util.BluesDateTime import BluesDateTime

class Behavior(Executor):
  def __init__(self,model:Model,browser:Browser=None)->None:
    super().__init__()

    self._validate(model.config)

    self._model = model 
    self._meta = model.meta
    self._bizdata = model.bizdata
    self._config = model.config
    self._browser = browser
    
  def execute(self)->STDOut:
    skip = self._config.get('skip',False)
    if skip:
      return STDOut(200,f'skip the behavior, the skip is True')

    self._wait('wait_before') 
    output:STDOut = self._invoke()
    self._wait('wait_after')
    return output
  
  @abstractmethod
  def _invoke(self):
    pass
  
  def _wait(self,key:str):
    wait_time:WaitTime = self._config.get(key)
    if not wait_time:
      return
    
    if isinstance(wait_time,list):
      if len(wait_time) == 2:
        second = wait_time[0]
        title = wait_time[1]
      elif len(wait_time) == 3:
        second = random.randint(wait_time[0],wait_time[1])
        title = wait_time[2]
    else:
      second = wait_time
      title = key
      
    BluesDateTime.count_down({
      'duration':second,
      'title':title
    })
  
  def _validate(self,instance:str):
    tpl_path = 'except.input.behavior'
    stat,message = SchemaValidator.validate_with_template(instance,tpl_path)
    if not stat:
      raise ValueError(f'behavior config error: {message}')

  def _get_kwargs(self,keys:list[str],config=None)->dict:
    '''
    Extract specified keys from configuration dictionary
    @param {list[str]} keys: list of keys to extract from config
    @param {dict} config: optional config dict to merge with self._config (config takes precedence)
    @return {dict}: dictionary containing only the specified keys and their values
    '''
    conf = {**self._config,**config} if config else self._config
    # must remove the attr that value is None
    key_conf = {}
    for key in keys:
      if key in conf:
        key_conf[key] = conf.get(key)
    return key_conf

  def _to_be_clickable(self)->bool:

    kwargs = self._get_kwargs(['loc_or_elem','timeout'])
    # scroll the element into view, avoid the element is covered by other elements
    return self._browser.waiter.ec.to_be_clickable(**kwargs)

  def _to_be_visible(self)->bool:
    kwargs = self._get_kwargs(['loc_or_elem','timeout'])
    return self._browser.waiter.ec.to_be_visible(**kwargs)

  def _to_be_presence(self)->bool:
    kwargs = self._get_kwargs(['loc_or_elem','timeout'])
    return self._browser.waiter.ec.to_be_presence(**kwargs)

  def _scroll(self):
    kwargs = self._get_kwargs(['loc_or_elem','parent_loc_or_elem'])
    scroll = self._config.get('scroll',False)
    if scroll:
      self._browser.action.wheel.scroll_element_to_center(**kwargs)
      time.sleep(0.1)

  def _after_invoked(self,value:any)->any:
    # 在getter后调用，用于处理获取到的值
    hook_defs:list[dict[str,any]] = self._config.get('after_invoked')
    if not hook_defs:
      return value
    
    options = {'value':value}
    BehaviorHook(hook_defs,options).execute()
    return options.get('value')


from blues_lib.hook.FuncHandler import FuncHandler
from blues_lib.type.output.STDOut import STDOut

class CommandFuncHandler(FuncHandler):
  
  def execute(self)->any:
    ti:any = self._options.get('ti')
    stdout:STDOut = self._options.get('stdout')
    self._func(ti,stdout)



from blues_lib.type.factory.Factory import Factory
from blues_lib.model.Model import Model

from blues_lib.behavior.unit.Row import Row
from blues_lib.behavior.unit.Richtext import Richtext

from blues_lib.behavior.chain.LightChain import LightChain

class BhvUnitFactory(Factory):
  def __init__(self,model:Model,browser=None):
    self._model = model
    self._browser = browser

  def create_row(self):
    return Row(self._model,self._browser)
  
  def create_richtext(self):
    return Richtext(self._model,self._browser)
  
  def create_light_chain(self):
    return LightChain(self._model,self._browser)



from abc import abstractmethod
from blues_lib.type.output.STDOut import STDOut
from blues_lib.behavior.Behavior import Behavior
from blues_lib.deco.BehaviorSTDOutLog import BehaviorSTDOutLog
from blues_lib.util.BluesDateTime import BluesDateTime
from blues_lib.types.common import IntervalTime,WaitTime

class Trigger(Behavior):

  @BehaviorSTDOutLog()
  def _invoke(self)->STDOut:
    value = None
    try:
      value = self._trigger()
      value = self._after_invoked(value)
      return STDOut(200,'ok',value)
    except Exception as e:
      return STDOut(500,str(e),value)

  @abstractmethod
  def _trigger(self)->any:
    pass
  
  def _wait_interval(self,index:int):
    defualt_value:int = 1
    interval:IntervalTime = self._config.get('interval',defualt_value)
    wait_time:int = 0
    title:str = f'wait to click {index+2}th element'

    if isinstance(interval,list):
      if 0 <= index < len(interval):
        item:WaitTime = interval[index]
        if isinstance(item,list):
          wait_time = item[0]
          title = item[1] or title
        else:
          wait_time = item
      else:
        wait_time = defualt_value
    else:
      wait_time = interval

    if wait_time > 0:
      BluesDateTime.count_down({
        'duration':wait_time,
        'title':title
      })


from .ChromeStdArgs import ChromeStdArgs   
from .ChromeExpArgs import ChromeExpArgs   
from .ChromeExtArgs import ChromeExtArgs   
from .ChromeCDPArgs import ChromeCDPArgs
from .ChromeSelArgs import ChromeSelArgs

class ChromeArgsFactory:
  
  def __init__(self,std_args=None,exp_args=None,cdp_args=None,sel_args=None,ext_args=None):
    self.__std_args = std_args or []
    self.__exp_args = exp_args or {}
    self.__cdp_args = cdp_args or {}
    self.__sel_args = sel_args or {}
    self.__ext_args = ext_args or {}

  def create(self)->dict:
    std_args = ChromeStdArgs(self.__std_args).get()
    exp_args = ChromeExpArgs(self.__exp_args).get()
    cdp_args = ChromeCDPArgs(self.__cdp_args).get()
    sel_args = ChromeSelArgs(self.__sel_args).get()
    ext_args = ChromeExtArgs(self.__ext_args).get()

    # connet to a exist browser, don't support experimental options setting 需要在Driver创建时设置，此种情况下仅连接 
    if sel_args.get('debugger_address'):
      std_args = None # has no effect
      exp_args = None # must remove
    return {
      'std':std_args,
      'exp':exp_args,
      'cdp':cdp_args,
      'sel':sel_args,
      'ext':ext_args,
    }

  def format(self)->dict:
    # 只返回结构，不设置任何默认配置
    return {
      'std':self.__std_args,
      'exp':self.__exp_args,
      'cdp':self.__cdp_args,
      'sel':self.__sel_args,
      'ext':self.__ext_args,
    }


from os import path
from urllib.parse import urlparse
from tld import get_fld
from blues_lib.util.BluesAlgorithm import BluesAlgorithm

class BluesURL():
  
  @classmethod
  def is_http_url(cls,url: str) -> bool:
    try:
      result = urlparse(url)
      # 检查 scheme 是否为 http 或 https，并且 netloc（域名）不为空
      return result.scheme in ('http', 'https') and bool(result.netloc)
    except ValueError:
      return False

  @classmethod
  def get_domain(cls,url):
    parsed_url = urlparse(url)
    return parsed_url.netloc

  @classmethod
  def get_main_domain(cls,url):
    return get_fld(url)
  
  @classmethod
  def get_abs_path(cls,abs_file,child_dir='',file_name=''):
    '''
    获取当前文件所在目录（或子目录下）的文件或目录的绝对路径
    @param {str} abs_file 当前文件绝对路径
    @param {str]} suffix 基于当前目录向后拼接的目录
    '''
    cur_path = path.dirname(abs_file)
    return path.join(cur_path,child_dir,file_name)
  
  @classmethod
  def get_file_name(cls,url:str,name:str='',with_extension:bool=True):
    parsed_url = urlparse(url)
    file_path = path.basename(parsed_url.path)
    file_name = path.splitext(file_path)[0] if not name else name
    extend_name = path.splitext(file_path)[1]

    if len(file_name)>50:
      file_name = BluesAlgorithm.md5(file_name)

    if with_extension:
      return file_name+extend_name
    else:
      return file_name

  @classmethod
  def get_extend_name(cls,file_name):
    '''
    Return the file's textend name, contains the .
    '''
    return path.splitext(file_name)[1]

  @classmethod
  def get_file_path(cls,dir_path,file_name):
    return path.join(dir_path,file_name)

  @classmethod
  def get_file_dir(cls,file_path):
    '''
    Get the file path witout file name
    '''
    return path.dirname(file_path)

  @classmethod
  def rename_extend_name(cls,file_path,extend_name):
    if not extend_name:
      return file_path

    file_dir = cls.get_file_dir(file_path)
    pure_file_name = cls.get_file_name(file_path,False)
    new_file_name = '%s.%s' % (pure_file_name,extend_name)
    return path.join(file_dir,new_file_name)
    



from copy import deepcopy
from abc import ABC,abstractmethod
from blues_lib.command.NodeCommand import NodeCommand
from blues_lib.type.output.STDOut import STDOut
from blues_lib.flow.Flow import Flow
from blues_lib.deco.CommandSTDOutLog import CommandSTDOutLog

class FlowCommand(NodeCommand,ABC):

  NAME = None

  def _invoke(self)->STDOut:
    return self._run()

  @abstractmethod
  def _run(self)->STDOut: 
    pass
    
  @CommandSTDOutLog('flow',max_chars=100)
  def _run_flow(self,dag_def:dict,dag_biz:list[dict])->STDOut:
    flow = self._get_flow(dag_def,dag_biz)
    stdout:STDOut = flow.execute() 
    if stdout.code == 200:
      return flow.last_task_stdout
    else:
      return stdout

  def _get_flow(self,dag_def:dict,dag_biz:list[dict])->Flow:
    # avoid circular import
    from blues_lib.flow.Flow import Flow
    flow = Flow(dag_def,dag_biz)
    if not flow:
      raise Exception(f'{self.NAME} Failed to create flow')
    return flow



from typing import Any

class ScriptExecutor:

  @classmethod
  def execute(cls,script:str,input:Any|None=None)->Any:
    if not script:
      return input

    try:
      if script.strip().startswith("lambda"):
        return cls._execute_lambda(script,input)
      elif script.strip().startswith("def func"):
        return cls._execute_func(script,input)
      else:
        raise ValueError("脚本字符串仅支持lambda函数（以lambda开头）或多行函数（以def func开头）")
    except SyntaxError as e:
      raise SyntaxError(f"脚本语法错误：{e}\n错误脚本：{script}") from e
    except NameError as e:
      raise NameError(f"脚本引用未定义的变量/函数：{e}\n错误脚本：{script}") from e
    except Exception as e:
      raise Exception(f"脚本执行失败：{e}\n错误脚本：{script}") from e
    
  @classmethod
  def _execute_lambda(cls,script:str,input:Any|None=None)->Any:
    func = eval(script,globals={}, locals={})
    return func(input)

  @classmethod
  def _execute_func(cls,script:str,input:Any|None=None)->Any:
    local_ns = {}
    exec(script, globals={}, locals=local_ns)
    func = local_ns["func"]
    return func(input)



from .Finder import Finder
from .XPathFinder import XPathFinder

from .Info import Info
from .State import State
from .Select import Select
from .Choice import Choice
from .File import File
from .Image import Image
from .Shot import Shot
from .Input import Input
from .Popup import Popup

class Element():

  def __init__(self,driver):
    self.finder = Finder(driver)
    self.xpath_finder = XPathFinder(driver)
    self.info = Info(driver)
    self.state = State(driver)
    self.select = Select(driver)
    self.choice = Choice(driver)
    self.file = File(driver)
    self.image = Image(driver)
    self.shot = Shot(driver)
    self.input = Input(driver)
    self.popup = Popup(driver)



from typing import Union
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.shadowroot import ShadowRoot
from blues_lib.sele.element.Locator import Locator  
from selenium.webdriver.support.relative_locator import locate_with
from .deco.FinderDeco import FinderDeco   

# 联合类型：直接包含所有支持元素查找的具体类型（适用于不需要泛型的场景）
SearchContext = Union[WebDriver, WebElement, ShadowRoot]

class Finder():
  def __init__(self,driver):
    self._driver = driver
  
  # === part 1:  geneal === #
  @FinderDeco('find')
  def find(self,loc_or_elem,parent_loc_or_elem=None)->WebElement|None:
    '''
    Get the first element in the DOM that matches with the provided locator.
    Parameter:
      loc_or_elem {str|WebElement} : the target element's css selector or WebElement - element_or_selector
      parent_loc_or_elem {str|WebElement} : the parent element's css selector or WebElement - element_or_selector
        - By default: the parent is the driver (the entire DOM)
    @returns {WebElement|None}
    '''
    if not loc_or_elem:
      return None

    if isinstance(loc_or_elem,WebElement):
      return loc_or_elem

    context: SearchContext = self.get_context(parent_loc_or_elem)
    return self._find(loc_or_elem,context)

  @FinderDeco('find_all')
  def find_all(self,loc_or_elem,parent_loc_or_elem=None)->list[WebElement]|None:
    '''
    Get all elements in the DOM that matches with the provided locator.
    loc_or_elem {str|WebElement} : the target element's css selector or WebElement
    parent_loc_or_elem {str|WebElement} : the parent element's css selector or WebElement
      - By default: the parent is the driver {WEbDriver} (the entire DOM)
    @returns {list<WebElement>}
    '''
    if not loc_or_elem:
      return None
    
    if isinstance(loc_or_elem,WebElement):
      return [loc_or_elem]
    
    context: SearchContext = self.get_context(parent_loc_or_elem)
    return self._find_all(loc_or_elem,context)

  # === part 2:  get shadow element === #
  @FinderDeco('find_shadow')
  def find_shadow(self,loc_or_elem,parent_loc_or_elem)->WebElement|None:
    '''
    Find the element in shadow root
    Parameter:
      loc_or_elem {str} : css selector of the element in the shadow root
      parent_loc_or_elem {str|WebElement} : the element contains the shadow root
    '''
    if not loc_or_elem:
      return None
    
    context: SearchContext|None = self._get_shadow_context(parent_loc_or_elem)
    if not context:
      return None

    return self._find(loc_or_elem,context)

  @FinderDeco('find_all_shadow')
  def find_all_shadow(self,loc_or_elem,parent_loc_or_elem)->list[WebElement]|None:
    '''
    Find all elements in shadow root
    Parameter:
      loc_or_elem {str} : css selector of the element in the shadow root
      parent_loc_or_elem {str|WebElement} : the element contains the shadow root
    '''
    if not loc_or_elem:
      return None
    
    context: SearchContext|None = self._get_shadow_context(parent_loc_or_elem)
    if not context:
      return None

    return self._find_all(loc_or_elem,context)

  # === part 4:  get element by other element postion === #
  @FinderDeco('find_above')
  def find_above(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->WebElement|None:
    '''
    Find the target element above the anchor element
    - 目标元素的底部边缘必须位于锚点元素的顶部边缘之上（即目标元素整体在锚点元素的正上方或斜上方）。
    - 无论目标元素和锚点元素在 HTML 结构中是否为父子关系、兄弟关系，只要视觉上满足 “目标在锚点上方”，就能被匹配。
    - 如果存在多个符合 “在锚点上方” 条件的元素，above() 会优先返回与锚点元素水平重叠最多的那个（即最 “正上方” 的元素）。
    - 只要目标元素底部边缘在锚点元素顶部边缘之上（无论两者垂直距离多远，哪怕间隔整个屏幕），都符合 above() 条件。
    Parameter:
      target_CS {str} : the target element's selector, general is a tag selector
      anchor_CS {str} : the anchor element's selector
    Returns:
      {WebElement]
    '''
    return self.find_relative('above',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_above_all(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->list[WebElement]|None:
    return self.find_relative_all('above',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_below(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->WebElement|None:
    return self.find_relative('below',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_below_all(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->list[WebElement]|None:
    return self.find_relative_all('below',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_left(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->WebElement|None:
    return self.find_relative('left',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_left_all(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->list[WebElement]|None:
    return self.find_relative_all('left',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_right(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->list[WebElement]|None:
    return self.find_relative('right',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_right_all(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->list[WebElement]|None:
    return self.find_relative_all('right',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_near(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->list[WebElement]|None:
    # you can use the near method to identify an element that is at most 50px away from the provided locator
    return self.find_relative('near',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_near_all(self,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->list[WebElement]|None:
    return self.find_relative_all('near',loc_or_elem,anchor_CS_WE,parent_loc_or_elem)

  def find_relative(self,position:str,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->WebElement|None:
    if not loc_or_elem or not anchor_CS_WE:
      return None

    locator = self._get_relative_locator(position,loc_or_elem,anchor_CS_WE)
    context = self.get_context(parent_loc_or_elem)
    return context.find_element(locator)

  def find_relative_all(self,position:str,loc_or_elem,anchor_CS_WE,parent_loc_or_elem=None)->list[WebElement]|None:
    if not loc_or_elem or not anchor_CS_WE:
      return None
    
    locator = self._get_relative_locator(position,loc_or_elem,anchor_CS_WE)
    context = self.get_context(parent_loc_or_elem)
    return context.find_elements(locator)

  # === appendix:  private methods === #
  def _get_relative_locator(self,position:str,loc_or_elem,anchor_CS_WE):
    target_locator:list[str] = Locator.get(loc_or_elem)
    anchor_ele = self.find(anchor_CS_WE)
    if not anchor_ele:
      return None

    if position == 'above':
      return locate_with(*target_locator).above(anchor_ele)
    elif position == 'below':
      return locate_with(*target_locator).below(anchor_ele)
    elif position == 'left':
      return locate_with(*target_locator).to_left_of(anchor_ele)
    elif position == 'right':
      return locate_with(*target_locator).to_right_of(anchor_ele)
    elif position == 'near':
      return locate_with(*target_locator).near(anchor_ele)
    
  def get_parent(self,parent_loc_or_elem)->WebElement|None:
    if not parent_loc_or_elem:
      return None

    # can't use type or undetected_chromedriver can not match
    if isinstance(parent_loc_or_elem,WebElement):
      return parent_loc_or_elem

    return self._find(parent_loc_or_elem,self._driver)
  
  def get_context(self,parent_loc_or_elem)->SearchContext:
    parent: WebElement|None = self.get_parent(parent_loc_or_elem)
    return parent or self._driver

  def _get_shadow_context(self,parent_loc_or_elem)->SearchContext|None:
    parent: WebElement|None = self.get_parent(parent_loc_or_elem)
    return parent.shadow_root if parent else None

  def _find(self,loc_or_elem:str,context:SearchContext)->WebElement|None:
    try:
      locator = Locator.get(loc_or_elem)
      return context.find_element(*locator)
    except Exception as e:
      return None

  def _find_all(self,loc_or_elem:str,context:SearchContext)->list[WebElement]|None:
    locator = Locator.get(loc_or_elem)
    return context.find_elements(*locator) or None



import sys,os,re
from abc import ABC,abstractmethod

from blues_lib.schema.reader.ifeng.IFengSchemaFactory import IFengSchemaFactory
from blues_lib.schema.reader.thepaper.ThePaperSchemaFactory import ThePaperSchemaFactory
from blues_lib.spider.MaterialSpider import MaterialSpider    
from blues_lib.publisher.StandardPublisher import StandardPublisher
from blues_lib.util.BluesConsole import BluesConsole

class NAPS(ABC):
  
  CHANNEL = ''

  '''
  1. Crawl a materail
  2. Login the publish page
  3. Publish
  4. Set published log
  '''
  def __init__(self):
    # {PublishPlan} current_quota current_total
    self._plan = self._get_plan()
  
  def execute(self):
    if self.__should_prevent():
      return

    self.spide()
    self.publish()
  
  def publish(self):
    if self.__should_prevent():
      return

    publisher = self.__get_publisher()
    publisher.publish()

  def prepublish(self):
    if self.__should_prevent():
      return

    publisher = self.__get_publisher()
    publisher.prepublish()

  def __get_publisher(self):
    loginer = self._get_loginer()
    models = self._get_models()
    return StandardPublisher(models,loginer)

  def __should_prevent(self):
    if self._plan.current_total<=0:
      BluesConsole.error('[%s] The limit for the day has been used up' % self.CHANNEL)
      return True
    else:
      return False

  @abstractmethod
  def _get_plan(self):
    pass

  @abstractmethod
  def _get_loginer(self):
    pass

  @abstractmethod
  def _get_models(self):
    pass

  def spide(self):
    '''
    Crawl a material
    Return:
      {bool}
    '''
    factory = ThePaperSchemaFactory()
    schema1 = factory.create_news('intl')

    factory = IFengSchemaFactory()
    schema2 = factory.create_tech_news()
    schema3 = factory.create_tech_outpost()
    schema4 = factory.create_hot_news()

    schemas = [schema2,schema3,schema4,schema1]

    spider = MaterialSpider(schemas,self._plan.current_total)
    return spider.spide()
 





import sys,os,re
from typing import Any

from blues_lib.behavior.Bean import Bean

class CSS(Bean):

  def _get(self)->str:
    kwargs = self._get_kwargs(['loc_or_elem','key','parent_loc_or_elem','timeout'])
    return self._browser.element.info.get_css(**kwargs)
  
  def _set(self):
    selector = self._config.get('loc_or_elem')
    parent_selector = self._config.get('parent_loc_or_elem')
    if entity := self._get_value_entity():
      if self._to_be_presence():
        return self._browser.script.javascript.css(selector,entity,parent_selector)



import re
from blues_lib.type.output.STDOut import STDOut
from blues_lib.util.BluesMailer import BluesMailer  
from blues_lib.command.NodeCommand import NodeCommand
from blues_lib.util.FileReader import FileReader
from blues_lib.namespace.CommandName import CommandName

class Email(NodeCommand):

  NAME = CommandName.Notifier.EMAIL

  def _invoke(self)->STDOut:
    payload = self.get_payload()
    mailer = BluesMailer.get_instance()
    return mailer.send(payload)
      
  def get_payload(self)->dict:
    entities:list[dict] = self._summary.get('entities',[])
    subject = self._get_subject(entities)
    paras = self._get_paras(entities)
    
    return {
      'subject':subject,
      'paras':paras,
      'addressee':['langcai10@dingtalk.com'], # send to multi addressee
      'addressee_name':'BluesLiu',
    }
    
  def _get_subject(self,entities:list[dict])->str:
    success_count = 0
    failure_count = 0
    for entity in entities:
      if entity.get('pub_stat') == 'success':
        success_count += 1
      else:
        failure_count += 1
    return f'Published : {success_count} success, {failure_count} failure'
    
  def _get_paras(self,entities:list[dict])->list[dict]:
    paras = []
    for entity in entities:
      paras.append({
        'type':'text',
        'value':f'{entity.get("mat_title")} - {entity.get("pub_stat")}',
      })
      paras.append({
        'type':'image',
        'value':entity.get('pub_shot'),
      })
    return paras

  def _get_log(self):
    file = self._logger.file
    separator = self._logger.separator
    content = FileReader.read(file)
    if content:
      # retain the latest one
      items = content.split(separator)
      non_empty_items = [item.strip() for item in items if item.strip()]
      content = non_empty_items[-1] if non_empty_items else content
      
      # break line
      content = content.replace('\n','<br/>')
      # dash line
      pattern = r'[-=]{10,}'
      content = re.sub(pattern, '----------', content)
    return content


from blues_lib.behavior.Trigger import Trigger

class UrlContains(Trigger):

  def _trigger(self)->bool:
    '''
    if the url contains the url slice in the wait time
    @returns {bool}
    '''
    url_slice = self._config.get('url_slice')
    wait_time = self._config.get('wait_time',3)
    return self._browser.waiter.ec.url_contains(url_slice,wait_time)



class HoconTemplate:
  
  @classmethod
  def include(cls,template:dict)->None:
    '''
    @description: replace the node value with the include file value
    @syntax: "%{include schema.except.stdout.200_list}"
      - update the dict directly
      - only support built-in file path
    @param {dict} config template dict
    @returns {None}
    '''
    return cls(**template)



import re
from selenium.webdriver.support.select import Select as DriverSelect
from selenium.webdriver.remote.webelement import WebElement
from .deco.SelectDeco import SelectDeco
from .deco.SelectKeyDeco import SelectKeyDeco

from blues_lib.sele.waiter.Querier import Querier  
from blues_lib.sele.element.Info import Info  

# 提供下拉选择相关功能
class Select():
  '''
  That this class only works for HTML elements select and option. 
  It is possible to design drop-downs with JavaScript overlays using div or li, and this class will not work for those
  '''
 
  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver,1)
    self.__info = Info(driver)
  
  def get(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Return the Select instance
    Parameter:
      loc_or_elem,parent_loc_or_elem {str|WebElement} : the html select element cess selector or web element
    Returns:
      {Select}
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return None
    return DriverSelect(web_element)
  
  @SelectDeco('get_options')
  def get_options(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    @description 获取select选项/选中对象list
    @param {str} loc_or_elem,parent_loc_or_elem
    @returns {list<dict>}
      [{'value': 'volvo', 'label': '沃尔沃'}]
    '''
    options = self.__get_options(loc_or_elem,parent_loc_or_elem,timeout)
    return self.__get_option_items(options)
      
  @SelectDeco('geted_options')
  def get_selected_options(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    @returns {list<dict>}
      [{'value': 'volvo', 'label': '沃尔沃'}]
    '''
    options = self.__get_options(loc_or_elem,'all_selected_options',parent_loc_or_elem,timeout)
    return self.__get_option_items(options)

  @SelectDeco('get_frist_selected_option')
  def get_first_selected_option(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    @returns {dict}
      {'value': 'volvo', 'label': '沃尔沃'}
    '''
    options = self.__get_options(loc_or_elem,'first_selected_option',parent_loc_or_elem,timeout)
    return self.__get_option_items(options)

  def __get_options(self,loc_or_elem,stat=None,parent_loc_or_elem=None,timeout=5):
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return None
    if stat == 'all_selected_options':
      return select.all_selected_options
    elif stat == 'first_selected_option':
      return select.first_selected_option
    else:
      return select.options

  def __get_option_items(self,option_elements):
    if not option_elements:
      return None
    
    # first selected option
    if type(option_elements) == WebElement:
      return {
        'value':option_elements.get_attribute('value'),
        'text':option_elements.text,
      }

    items = []
    for option in option_elements:
      items.append({
        'value':option.get_attribute('value'),
        'text':option.text,
      })
    return items

  @SelectKeyDeco('select_by_index')
  def select_by_index(self,loc_or_elem,index,parent_loc_or_elem=None,timeout=5):
    '''
    Select a option by index
    Parameter:
      index {int} : the option's index
      loc_or_elem {str|WebElement} : the select element
      parent_loc_or_elem {str|WebElement} : the select's parent element
    Returns:
      {None}
    '''
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return False

    indexs = index if type(index)==list else [index]
    for idx in indexs:
      select.select_by_index(idx)

    return len(indexs)
  
  @SelectKeyDeco('select_by_value')
  def select_by_value(self,loc_or_elem,value,parent_loc_or_elem=None,timeout=5):
    '''
    Selet multi values
    '''
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return False

    values = value if type(value)==list else [value]
    for val in values:
      select.select_by_value(val)

    return len(values)

  @SelectKeyDeco('select_by_value_or_text')
  def select_by_value_or_text(self,loc_or_elem,value,parent_loc_or_elem=None,timeout=5):
    '''
    Selet multi values
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return None

    select = self.get(web_element)

    values = value if type(value)==list else [value]
    
    first_option_html = self.__info.get_outer_html('option',web_element)

    if re.search(r'value\s*=',first_option_html):  
      for val in values:
        select.select_by_value(val)
    else:
      for val in values:
        select.select_by_visible_text(val)

    return len(values)

  @SelectKeyDeco('select_by_text')
  def select_by_text(self,loc_or_elem,text,parent_loc_or_elem=None,timeout=5):
    '''
    Returns:
      {None}
    '''
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return False

    texts = text if type(text)==list else [text]
    for txt in texts:
      select.select_by_visible_text(txt)

    return len(texts)

  @SelectDeco('deselect_all')
  def deselect_all(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    You may only deselect all options of a multi-select
    Returns:
      {None}
    '''
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return False
    if not select.is_multiple:
      return False
    return select.deselect_all()
    
  @SelectKeyDeco('deselect_by_index')
  def deselect_by_index(self,loc_or_elem,index,parent_loc_or_elem=None,timeout=5):
    '''
    Returns:
      {None}
    '''
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return False
    if not select.is_multiple:
      return False

    indexs = index if type(index)==list else [index]
    for idx in indexs:
      select.deselect_by_index(idx)

    return len(indexs)
  
  @SelectKeyDeco('deselect_by_value')
  def deselect_by_value(self,loc_or_elem,value,parent_loc_or_elem=None,timeout=5):
    '''
    Returns:
      {None}
    '''
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return False
    if not select.is_multiple:
      return False

    values = value if type(value)==list else [value]
    for val in values:
      select.deselect_by_value(val)

    return len(values)

  @SelectKeyDeco('deselect_by_text')
  def deselect_by_text(self,loc_or_elem,text,parent_loc_or_elem=None,timeout=5):
    '''
    Returns:
      {None}
    '''
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return False
    if not select.is_multiple:
      return False

    texts = text if type(text)==list else [text]
    for txt in texts:
      select.deselect_by_visible_text(txt)

    return len(texts)

  @SelectDeco('is_multiple')
  def is_multiple(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Returns:
      {true|None}
    '''
    select = self.get(loc_or_elem,parent_loc_or_elem,timeout)
    if not select:
      return False
    return select.is_multiple






from selenium.webdriver.chrome.options import Options
from blues_lib.sele.browser.driver.args.ChromeArgsFactory import ChromeArgsFactory

class OptionsCreator():
  '''
  创建自定义options
  '''
  
  def __init__(self,
    std_args=None, # {list} standard args
    exp_args=None, # {dict} experimental args
    cdp_args=None, # {dict} chrome devtools protocal args
    sel_args=None, # {dict} selenium args
    ext_args=None, # {dict} extension args
    undetected=True, # {bool} 是否使用 undetected_chromedriver
    ):
    self._std_args = std_args or []
    self._exp_args = exp_args or {}
    self._cdp_args = cdp_args or {}
    self._sel_args = sel_args or {}
    self._ext_args = ext_args or {}
    self._undetected = undetected
    self._arg_dict = None
    
  def create(self):
    factory = ChromeArgsFactory(self._std_args,self._exp_args,self._cdp_args,self._sel_args,self._ext_args)
    self._arg_dict = factory.format() if self._undetected else factory.create()

    options = Options()
    self._set_std(options)
    self._set_exp(options)
    self._set_ext(options)
    self._set_sel(options)
    return options
  
  def is_undetected_driver(self,driver):
    # 检查类名是否包含 UDC 特征
    class_module = getattr(driver.__class__, '__module__', '')
    driver_name = getattr(driver, '__name__', '')
    name = "undetected_chromedriver"
    return name in class_module or name in driver_name

  def _set_std(self,options):
    if args:= self._arg_dict.get('std'):
      for value in args:
        options.add_argument(value)

  def _set_exp(self,options):
    if args:= self._arg_dict.get('exp'):
      for key,value in args.items():
        options.add_experimental_option(key,value)

  def _set_ext(self,options):
    if args:= self._arg_dict.get('ext'):
      for value in args:
        options.add_extension(value)

  def _set_sel(self,options):
    if args:= self._arg_dict.get('sel'):
      for key,value in args.items():
        # just set as the options's attr
        setattr(options,key,value)
        
  def set_cdp(self,driver):
    # 此方法要在driver创建后调用,才能生效
    if args:= self._arg_dict.get('cdp'):
      for key,value in args.items():
        driver.execute_cdp_cmd(key,value)


from abc import ABC
import inspect
from blues_lib.namespace.NSEnum import NSEnum

class EnumNS(ABC):
  @classmethod
  def from_value(cls,value:str)->NSEnum|None:
    label = value.replace('-','.')
    for name,subclass in inspect.getmembers(cls):
      if inspect.isclass(subclass) and issubclass(subclass,NSEnum):
        member = subclass.from_value(label)
        if member is not None:
          return member
    return None



class BluesList(list):

  @classmethod 
  def get_length(cls,strings: list[str]) -> int:
    """
    * Calculates the total length of all valid strings in the provided list.
    * Non-string elements are silently ignored.
    * 
    * @param strings List of objects, strings are processed for length
    * @return Total length of all valid strings combined
    """
    total_length = 0
    for element in strings:
      if isinstance(element, str):
        total_length += len(element)
    return total_length


import re
from PIL import Image
from blues_lib.util.BluesURL import BluesURL

class BluesImage():
  
  '''
  BaiJia news:412 * 545
  '''
  MIN_WIDTH = 650
  MIN_HEIGHT = 650
  
  @classmethod
  def convert_type(cls,image_path,image_type='JPEG'):
    '''
    Convert the image's type

    Parameter:
      image_path {str} : the local image dir path
      image_type {str} : the aim image type

    Return:
      {str} : the converted image path
        If converted failure, return empty string
    '''
    file_name = BluesURL.get_file_name(image_path)
    jpeg_pattern = r'\.jpe?g$'
    if re.search(jpeg_pattern,file_name,flags=re.IGNORECASE):
      return image_path
    else:
      new_path = BluesURL.rename_extend_name(image_path,'jpg')
      try:
        image = Image.open(image_path)
        image.convert('RGB').save(new_path, image_type)
        image.close()
        return new_path
      except Exception as e:
        return ''

  @classmethod
  def convert_size(cls,image_path,size=None):
    '''
    Convert the image's size

    Parameter:
      image_path {str} : the local image dir path
      size {tuple|list} : the aim image size ,such as [100,200] [100,] [,200]

    Return:
      {bool} : the convet status
    '''
    if size:
      return cls.resize(image_path,size)

    image = Image.open(image_path)
    width,height = image.size
    image.close()
    if width>= cls.MIN_WIDTH and height>=cls.MIN_HEIGHT:
      return 
    
    size = cls.get_min_size(width,height)
    return cls.resize(image_path,size)

  @classmethod
  def get_min_size(cls,width,height):

    new_width = width
    new_height = height

    # Make sure both the width and height large than the MIN value
    if width>= cls.MIN_WIDTH and height<cls.MIN_HEIGHT:
      new_height = cls.MIN_HEIGHT
      new_width = int(width*new_height/height) 

    if width<cls.MIN_WIDTH and height>=cls.MIN_HEIGHT:
      new_width = cls.MIN_WIDTH
      new_height = int(new_width*height/width) 

    if width<cls.MIN_WIDTH and height<cls.MIN_HEIGHT:
      if width/cls.MIN_WIDTH < height/cls.MIN_HEIGHT:
        new_width = cls.MIN_WIDTH
        new_height = int(new_width*height/width) 
      else:
        new_height = cls.MIN_HEIGHT
        new_width = int(width*new_height/height) 

    size =  (new_width,new_height)
    return size

  @classmethod
  def resize(cls,image_path,size):
    '''
    Resize the image
    Parameter:
      image_path {str} : the local image path
      size {tuple,list} : (width,height)
    '''
    try:
      image = Image.open(image_path)
      image.resize(size).save(image_path)
      image.close()
      return True
    except Exception as e:
      return False





import os
from blues_lib.model.BaseVarReplacer import BaseVarReplacer


class EnvVarReplacer(BaseVarReplacer):
  '''
  Environment variable replacer that replaces variable placeholders with environment variable values.
  Uses the same format as BaseVarReplacer but gets values from system environment variables.
  '''
  
  @classmethod
  def _get_raw_value(cls, var_name: str) -> str | None:
    '''
    Get value from system environment variables
    
    Args:
        var_name: Environment variable name
        
    Returns:
        Environment variable value or None if not found
    '''
    return os.environ.get(var_name)


from blues_lib.type.factory.Factory import Factory
from blues_lib.model.Model import Model

from blues_lib.behavior.keyboard.Keyboard import Keyboard

from blues_lib.behavior.bean.Length import Length
from blues_lib.behavior.bean.Attr import Attr
from blues_lib.behavior.bean.CSS import CSS
from blues_lib.behavior.bean.Choice import Choice
from blues_lib.behavior.bean.Checkbox import Checkbox

from blues_lib.behavior.bean.File import File
from blues_lib.behavior.bean.AuthCodeInput import AuthCodeInput
from blues_lib.behavior.bean.Select import Select
from blues_lib.behavior.bean.Screenshot import Screenshot
from blues_lib.behavior.bean.Text import Text
from blues_lib.behavior.bean.Html import Html
from blues_lib.behavior.bean.PageSource import PageSource
from blues_lib.behavior.bean.Value import Value
from blues_lib.behavior.bean.FileCookie import FileCookie
from blues_lib.behavior.bean.BrCookie import BrCookie
from blues_lib.behavior.bean.Url import Url
from blues_lib.behavior.bean.Copy import Copy

from blues_lib.behavior.event.Click import Click
from blues_lib.behavior.event.Framein import Framein
from blues_lib.behavior.event.Frameout import Frameout
from blues_lib.behavior.event.Hover import Hover
from blues_lib.behavior.event.Remove import Remove
from blues_lib.behavior.event.Scroll import Scroll
from blues_lib.behavior.event.Open import Open
from blues_lib.behavior.event.Quit import Quit
from blues_lib.behavior.event.Close import Close
from blues_lib.behavior.event.Switch import Switch

from blues_lib.behavior.checker.ElePresents import ElePresents
from blues_lib.behavior.checker.EleAbsents import EleAbsents
from blues_lib.behavior.checker.EleInvisible import EleInvisible
from blues_lib.behavior.checker.UrlChanges import UrlChanges
from blues_lib.behavior.checker.UrlContains import UrlContains
from blues_lib.behavior.checker.UrlToBe import UrlToBe
from blues_lib.behavior.checker.UrlMatches import UrlMatches
from blues_lib.behavior.checker.ToBeClickable import ToBeClickable
from blues_lib.behavior.checker.ToBeSelected import ToBeSelected
from blues_lib.behavior.checker.ToBePresence import ToBePresence
from blues_lib.behavior.checker.ToBeVisible import ToBeVisible

from blues_lib.behavior.general.Wait import Wait
from blues_lib.behavior.general.Email import Email

from blues_lib.behavior.script.Javascript import Javascript
from blues_lib.behavior.script.Python import Python

from blues_lib.behavior.write.WriteText import WriteText
from blues_lib.behavior.write.WriteChar import WriteChar
from blues_lib.behavior.write.WritePara import WritePara
from blues_lib.behavior.write.ResetChar import ResetChar

class BhvFactory(Factory):
  def __init__(self,model:Model,browser=None):
    self._model = model
    self._browser = browser

  def create_keyboard(self):
    return Keyboard(self._model,self._browser)
  
  def create_length(self):
    return Length(self._model,self._browser)
  
  def create_attr(self):
    return Attr(self._model,self._browser)
  
  def create_css(self):
    return CSS(self._model,self._browser)
  
  def create_choice(self):
    return Choice(self._model,self._browser)
  
  def create_checkbox(self):
    return Checkbox(self._model,self._browser)
  
  def create_file(self):
    return File(self._model,self._browser)
  
  def create_auth_code_input(self):
    return AuthCodeInput(self._model,self._browser)
  
  def create_select(self):
    return Select(self._model,self._browser)
  
  def create_screenshot(self):
    return Screenshot(self._model,self._browser)
  
  def create_text(self):
    return Text(self._model,self._browser)
  
  def create_html(self):
    return Html(self._model,self._browser)
  
  def create_page_source(self):
    return PageSource(self._model,self._browser)
  
  def create_value(self):
    return Value(self._model,self._browser)
  
  def create_file_cookie(self):
    return FileCookie(self._model,self._browser)
  
  def create_br_cookie(self):
    return BrCookie(self._model,self._browser)
  
  def create_url(self):
    return Url(self._model,self._browser)
  
  def create_copy(self):
    return Copy(self._model,self._browser)
  
  def create_click(self):
    return Click(self._model,self._browser)
  
  def create_framein(self):
    return Framein(self._model,self._browser)
  
  def create_frameout(self):
    return Frameout(self._model,self._browser)
  
  def create_hover(self):
    return Hover(self._model,self._browser)
  
  def create_remove(self):
    return Remove(self._model,self._browser)
  
  def create_scroll(self):
    return Scroll(self._model,self._browser)
  
  def create_open(self):
    return Open(self._model,self._browser)
  
  def create_quit(self):
    return Quit(self._model,self._browser)
  
  def create_close(self):
    return Close(self._model,self._browser)
  
  def create_switch(self):
    return Switch(self._model,self._browser)
  
  def create_ele_presents(self):
    return ElePresents(self._model,self._browser)
  
  def create_ele_absents(self):
    return EleAbsents(self._model,self._browser)
  
  def create_ele_invisible(self):
    return EleInvisible(self._model,self._browser)

  def create_url_changes(self):
    return UrlChanges(self._model,self._browser)
    
  def create_url_contains(self):
    return UrlContains(self._model,self._browser)
  
  def create_url_to_be(self):
    return UrlToBe(self._model,self._browser)
  
  def create_url_matches(self):
    return UrlMatches(self._model,self._browser)
  
  def create_to_be_clickable(self):
    return ToBeClickable(self._model,self._browser)

  def create_to_be_selected(self):
    return ToBeSelected(self._model,self._browser)
  
  def create_to_be_presence(self):
    return ToBePresence(self._model,self._browser)
  
  def create_to_be_visible(self):
    return ToBeVisible(self._model,self._browser)
  
  def create_wait(self):
    return Wait(self._model)
  
  def create_email(self):
    return Email(self._model)
  
  def create_javascript(self):
    return Javascript(self._model,self._browser)
  
  def create_python(self):
    return Python(self._model,self._browser)
  
  def create_write_text(self):
    return WriteText(self._model,self._browser)

  def create_write_char(self):
    return WriteChar(self._model,self._browser)
  
  def create_write_para(self):
    return WritePara(self._model,self._browser)
  
  def create_reset_char(self):
    return ResetChar(self._model,self._browser)


class JSLoader():

  def load_script(self,url):
    '''
    @description : 导入线上/本地 js脚本，用异步执行，保证python等待资源加载完毕
    @param {str} url : the scirpt file's path
    '''
    callback = "var callback = arguments[arguments.length-1];"
    create_script = "var sele_script=document.createElement('script');sele_script.src='%s';" % url
    # 在onload中执行callback，通知python js脚本已结束
    onload_script = "sele_script.onload=function(){callback();console.log('%s loaded by the selenium!')};" % url
    append_script = "var head=document.getElementsByTagName('head')[0];head.appendChild(sele_script);"
    load_script = "%s %s %s %s" % (callback,create_script,onload_script,append_script)
    # wait the script loaded, Determine whether it is complete according to the dynamic insertion mark 'selenium-dynamic-script'
    return self.execute_async(load_script)

  def load_blues(self):
    js_script = BluesJavaScriptService.service_script
    self.execute(js_script)

  def load_plugin(self,plugins):
    '''
    @description : load relay plugin auto
    @param {dict} plugins 
      - key : the global var name
      - value : the plugin's cdn
    '''
    for plugin_var,plugin_cdn in plugins.items():
      print('===>',plugin_var,plugin_cdn)
      if self.is_var_available(plugin_var):
        continue
      self.load_script(plugin_cdn)



from blues_lib.schema.SchemaValidator import SchemaValidator
from blues_lib.type.output.STDOut import STDOut

class IOExcept():

  @classmethod
  def validate_dag(cls,def_def:dict)->None:
    prefix:str = f'{def_def.get("dag_id")} dag :: '
    cls.validate_with_template(prefix,def_def,'dag')
    
  @classmethod
  def validate_task(cls,task_def:dict)->None:
    prefix:str = f'{task_def.get("task_id")} [{task_def.get("command")}] task :: '
    cls.validate_with_template(prefix,task_def,'task')
    
  @classmethod
  def validate_task_input(cls,task_def:dict)->None:
    input_defs:list[dict]|None = task_def.get('input')
    if not input_defs:
      return

    prefix:str = f'{task_def.get("task_id")} [{task_def.get("command")}] input :: '
    cls.validate_with_template(prefix,input_defs,'input')
    
  @classmethod
  def validate_metadata(cls,metadata:dict):
    # the $id identify the metadata's schema dot path
    if not (tpl_path := metadata.get('$id')):
      return

    prefix:str = f'{tpl_path} :: '
    cls.validate_with_template(prefix,metadata,tpl_path)
    

  @classmethod
  def validate_except(cls,task_def:dict,stdout:STDOut|None)->None:
    schema = task_def.get('except')
    if not schema:
      return 

    prefix:str = f'{task_def.get("task_id")} [{task_def.get("command")}] except :: '
    if not stdout:
      raise ValueError(f'{prefix}stdout is None')

    instance = stdout.to_dict()
    if isinstance(schema,str):
      tpl_path = f'except.{schema}'
      cls.validate_with_template(prefix,instance,tpl_path)
    else:
      cls.validate(prefix,instance,schema)
    

  @classmethod
  def validate_with_template(cls,prefix:str,instance:any,tpl_path:str)->None:
    prefix =  prefix if prefix else f'{tpl_path} :: '
    stat,message = SchemaValidator.validate_with_template(instance,tpl_path)
    if not stat:
      raise ValueError(f'{prefix}{message}')
    
  @classmethod
  def validate(cls,prefix:str,instance:any,schema:dict)->None:
    prefix =  prefix if prefix else 'schema validate :: '
    stat,message = SchemaValidator.validate(instance,schema)
    if not stat:
      raise ValueError(f'{prefix}{message}')
    
  
    

    

    

    

    

    

    

    



from queue import Queue
import threading

class QueueFlow:
    def __init__(self, max_commands: int = 10):
        self.command_queue = Queue(maxsize=max_commands)  # Store commands
        self.worker_thread = threading.Thread(target=self._process_commands, daemon=True)
        self.worker_thread.start()

    def add_command(self, command: Command) -> None:
        """Add a command to the flow queue"""
        self.command_queue.put(command)  # Blocks if queue is full

    def _process_commands(self) -> None:
        """Worker loop to process commands sequentially"""
        while True:
            command = self.command_queue.get()  # Blocks if queue is empty
            try:
                command.execute()  # Run the command
            finally:
                self.command_queue.task_done()  # Mark command as processed

    def wait_for_completion(self) -> None:
        """Wait for all commands in the queue to finish"""
        self.command_queue.join()


import os
import platform
from dotenv import load_dotenv

class EnvLoader:
  """
  Environment variable loader, used to load environment variable configuration files in the project
  Supports loading common configuration files and system-specific configuration files

  os.environ 本质是 Python 进程启动时创建的一个全局字典，用于存储当前进程的所有环境变量（包括系统环境变量和 load_dotenv() 注入的 .env 配置）。
  这个字典的关键特性是：一旦在某个模块中被修改（比如 load_dotenv() 往里面添加了新键值对），所有其他模块通过 import os 访问到的 os.environ 都是同一个对象—— 就像一个 “共享的白板”，任何模块写的内容，其他模块都能看到。
  因此，你只需要在程序启动的最早阶段（主文件开头） 执行一次 load_dotenv()，就能让整个项目的所有模块共享这份配置。
  """
  
  # Private static fields for environment file names
  __BASE_ENV_FILE = ".env"
  __WINDOWS_ENV_FILE = ".env.windows"
  __MACOS_ENV_FILE = ".env.macos"
  
  # Static variable to track if environment has been loaded
  __is_loaded = False
  
  @classmethod
  def load(cls):
    """
    Load environment variable configuration files.
    First load .env file, then load system-specific configuration file based on current OS.
    System-specific configuration will override common configuration.
    
    @return {bool} - Whether at least one configuration file was successfully loaded
    """
    # Check if already loaded
    if cls.__is_loaded:
      print("Environment variables already loaded. Skipping reloading.")
      return True
      
    loaded = False
    
    # 1. Load common configuration
    if os.path.exists(cls.__BASE_ENV_FILE):
      load_dotenv(cls.__BASE_ENV_FILE)
      loaded = True
      print(f"Loaded base configuration file: {cls.__BASE_ENV_FILE}")
    else:
      print(f"Warning: Base configuration file {cls.__BASE_ENV_FILE} not found.")
    
    # 2. Dynamically detect OS and load system-specific configuration
    os_name = platform.system()
    if os_name == "Windows":
      env_file = cls.__WINDOWS_ENV_FILE
    elif os_name == "Darwin":  # macOS
      env_file = cls.__MACOS_ENV_FILE
    else:
      print(f"Warning: Unsupported operating system type {os_name}.")
      # Mark as loaded after successful execution
      cls.__is_loaded = True
      return loaded
    
    if os.path.exists(env_file):
      load_dotenv(env_file, override=True)  # override=True ensures system configuration can override base configuration
      loaded = True
      print(f"Loaded and overridden system configuration file: {env_file}")
    else:
      print(f"Warning: System configuration file {env_file} not found.")
    
    # Mark as loaded after successful execution
    cls.__is_loaded = True
    return loaded


from .Navi import Navi
from .Alert import Alert
from .Frame import Frame
from .Cookie import Cookie
from .Document import Document
from .Window import Window

class Interactor():

  def __init__(self,driver):
    self.navi = Navi(driver)
    self.alert = Alert(driver)
    self.frame = Frame(driver)
    self.cookie = Cookie(driver)
    self.document = Document(driver)
    self.window = Window(driver)



from abc import abstractmethod
from blues_lib.behavior.BhvExecutor import BhvExecutor
from blues_lib.type.executor.Executor import Executor
from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model
from blues_lib.sele.browser.Browser import Browser 
from blues_lib.namespace.CrawlerName import CrawlerName
from blues_lib.util.BluesDateTime import BluesDateTime

class Crawler(Executor):

  def __init__(self,model:Model,browser:Browser) -> None:
    '''
    @param model {Model} : the model of crawler
    @param browser {Browser} : the browser instance to use
    '''
    super().__init__()
    self._model:Model = model
    self._browser:Browser = browser

  def _setup(self):
    # model
    self._conf:dict = self._model.config 
    self._meta:dict = self._model.meta
    self._bizdata:dict = self._model.bizdata
    
    # summary
    self._summary_conf:dict = self._conf.get(CrawlerName.Field.SUMMARY.value,{})
    self._count :int = self._summary_conf.get(CrawlerName.Field.COUNT.value,-1)
    # by default, quit the browser after crawled
    self._quit :bool = self._summary_conf.get(CrawlerName.Field.QUIT.value,True)

    # hook
    self._before_crawled_conf = self._conf.get(CrawlerName.Field.BEFORE_CRAWLED.value,{})
    self._after_crawled_conf = self._conf.get(CrawlerName.Field.AFTER_CRAWLED.value,{})
    self._before_each_crawled_conf = self._conf.get(CrawlerName.Field.BEFORE_EACH_CRAWLED.value,{})
    self._after_each_crawled_conf = self._conf.get(CrawlerName.Field.AFTER_EACH_CRAWLED.value,{})
    
    # head crawler
    self._head_meta:dict = self._meta.get(CrawlerName.Field.HEAD_CRAWL.value)
    self._head_conf:dict = self._conf.get(CrawlerName.Field.HEAD_CRAWL.value)

    # foot crawler
    self._foot_meta:dict = self._meta.get(CrawlerName.Field.FOOT_CRAWL.value)
    self._foot_conf:dict = self._conf.get(CrawlerName.Field.FOOT_CRAWL.value)

  def execute(self)->STDOut: 
    
    # Template method: define the cal structure
    self._setup()
    
    self._before_crawled()
    self._head()
    output:STDOut = self._crawl()
    self._foot(output)
    self._after_crawled(output)

    self._slice(output)
    self._close()
    self._log(output)
    return output
  
  def _before_crawled(self):
    pass

  def _head(self)->any:
    # execute the head crawler
    if self._head_meta:
      # must pass the meta and bizdata, some behavior need to calculate the model
      model = Model(self._head_meta,self._bizdata)
      return self._invoke(model)

  @abstractmethod
  def _crawl(self)->STDOut:
    pass
  
  def _foot(self,output:STDOut)->any:
    # execute the head crawler
    if self._foot_meta:
      # must pass the meta and bizdata, some behavior need to calculate the model
      model = Model(self._foot_meta,self._bizdata)
      return self._invoke(model)
  
  def _invoke(self,model:Model)->STDOut:
    # execute a base crawler
    try:
      bhv = BhvExecutor(model,self._browser)
      stdout:STDOut = bhv.execute()
      if isinstance(stdout.data,dict):
        stdout.data = stdout.data.get(CrawlerName.Field.DATA.value)
      return stdout
    except Exception as e:
      message = f'[{self.NAME}] Failed to crawl - {e}'
      self._logger.error(message)
      return STDOut(500,message)
  
  def _after_crawled(self,output:STDOut):
    pass

  def _slice(self,output:STDOut):
    # slice the data by the config count

    if self._count==-1 or output.code!=200:
      return
    
    if not output.data or not isinstance(output.data,list):
      return

    output.data = output.data[:self._count]
  
  def _log(self,output:STDOut):
    if output.code != 200:
      message = f'[{self.NAME}] Failed to crawl - {output.message}'
      self._logger.error(message)
    else:
      message = f'[{self.NAME}] Managed to crawl'
      self._logger.info(message)

  def _open(self,url:str):
    self._browser.open(url)

  def _close(self):
    if self._quit and self._browser:
      print('----quit---')
      self._browser.quit()

  def _set_interval(self):
    interval = self._summary_conf.get(CrawlerName.Field.LOOP_INTERVAL.value) or 1
    BluesDateTime.count_down({
      "duration":interval,
      "title":"loop interval",
    })



from selenium.webdriver.remote.webelement import WebElement
class JSDocument():

  # == module 1: html or text == #
  def html(self,selector,html=None,parent_selector=''): 
    '''
    @description : get or set html
    @param {str} selector css选择器
    @param {str} html 写入文本或HTML
    @returns {str|None} return dom html when as getter
    '''
    if html==None:
      if parent_selector and type(parent_selector)==str:
        script = 'return document.querySelector(`%s %s`).innerHTML;' % (parent_selector,selector)
      elif isinstance(parent_selector, WebElement):
        script = 'return arguments[0].querySelector(`%s`).innerHTML;' % (selector)
      else:
        script = 'return document.querySelector(`%s`).innerHTML;' % (selector)
    else:
      if parent_selector and type(parent_selector)==str:
        script = 'document.querySelector(`%s %s`).innerHTML=`%s`;' % (parent_selector,selector,html)
      elif isinstance(parent_selector, WebElement):
        script = 'return arguments[0].querySelector(`%s`).innerHTML=`%s`;' % (selector,html)
      else:
        script = 'document.querySelector(`%s`).innerHTML=`%s`;' % (selector,html)
    return self.execute(script,parent_selector)
  
  def replace(self,selector,html=''):
    # remove all
    script = '''
      var web_elements = document.querySelectorAll(`%s`);
      var cal_count = 0;
      if(web_elements.length){
        for(var i=0; i<web_elements.length;i++){
          web_element = web_elements[i];
          cal_count+=1;
          web_element.outerHTML=`%s`;
        }
      }
      return cal_count;
    ''' % (selector,html)
    return self.execute(script)
  
  def text(self,selector,text=None,parent_selector=''):
    '''
    @description : get or set text
    @param {str} selector css选择器
    @param {str} text 
    @returns {str|None} return dom html when as getter
    '''
    if text==None:
      if parent_selector and type(parent_selector)==str:
        script = 'return document.querySelector(`%s %s`).innerText;' % (parent_selector,selector)
      elif isinstance(parent_selector, WebElement):
        script = 'return arguments[0].querySelector(`%s`).innerText;' % (selector)
      else:
        script = 'return document.querySelector(`%s`).innerText;' % (selector)
    else:
      if parent_selector and type(parent_selector)==str:
        script = 'document.querySelector(`%s %s`).innerText=`%s`;' % (parent_selector,selector,text)
      elif isinstance(parent_selector, WebElement):
        script = 'return arguments[0].querySelector(`%s`).innerText=`%s`;' % (selector,text)
      else:
        script = 'document.querySelector(`%s`).innerText=`%s`;' % (selector,text)
    return self.execute(script,parent_selector)

  def html_after(self,selector,html=''):
    '''
    @description PM04 追加元素内容
    @param {str} selector css选择器
    @param {str} html 写入文本或HTML
    '''
    script = '''
      var ele = document.querySelector(`%s`); 
      var html = ele.innerHTML;
      ele.innerHTML=html+`%s`;
    ''' % (selector,html)
    return self.execute(script)

  def text_after(self,selector,text):
    '''
    @description : append text 
    @param {str} selector css选择器
    @param {str} text 
    '''
    script = '''
      var ele = document.querySelector(`%s`); 
      var text = ele.innerText;
      ele.innerText=text+`%s`;
    ''' % (selector,text)
    return self.execute(script)

  # == module 2: remove and empty == #
  def remove(self,selector):
    '''
    @description : remove the html dom
    @param {str} selector : css selector
    '''
    return self.replace(selector,'')

  def empty(self,selector):
    '''
    @description : remove the html dom
    @param {str} selector : css selector
    '''
    return self.html(selector,'')

  # == module 3: value == #
  def write(self,selector,value=''):
    '''
    @description : set form element's valu
    '''
    script = '''
      document.querySelector(`%s`).value=`%s`;
    ''' % (selector,value)
    return self.execute(script)

  
  def write_after(self,selector,value=''):
    '''
    @description : set form controller's value
    '''
    script = '''
      var ele = document.querySelector(`%s`); 
      var value = ele.value;
      ele.value=value+`%s`;
    ''' % (selector,value)
    self.execute(script)

  def write_para(self,selector,texts,LF_count=1):
    '''
    Write paras to textarea
    '''
    value = ''
    if texts:
      for text in texts:
        value+=text+('\n'*LF_count)

    return self.write(selector,value)

  # == module 4: attr and css == #
  def attr(self,selector,key_or_attrs):
    '''
    @description : get or set element's attribute
    @param {str} selector : element's css selector
    @param {str|dict} key_or_attrs  
    @returns {str|None} : return the attr's value in gtter
    '''
    if type(key_or_attrs)==str:
      script = 'return document.querySelector(`%s`).getAttribute(`%s`);' % (selector,key_or_attrs)
      return self.execute(script)
    else:
      for key,value in key_or_attrs.items():
        script = 'document.querySelector(`%s`).setAttribute(`%s`,`%s`);' % (selector,key,value)
        self.execute(script)
   
  def css(self,selector,key_or_attrs,parent_selector=''):
    '''
    @description : get or set element's style
    @param {str} selector : element's css selector
    @param {str|dict} key_or_attrs  
    @returns {str|None} : return the css style's value in gtter
    '''
    if type(key_or_attrs)==str:
      # getter
      if parent_selector and type(parent_selector)==str:
        script = 'return document.querySelector(`%s %s`).style[`%s`];' % (parent_selector,selector,key_or_attrs)
      elif isinstance(parent_selector, WebElement):
        script = 'return arguments[0].querySelector(`%s`).style[`%s`];' % (selector,key_or_attrs)
      else:
        script = 'return document.querySelector(`%s`).style[`%s`];' % (selector,key_or_attrs)
      return self.execute(script,parent_selector)
    else:
      # setter
      for key,value in key_or_attrs.items():
        if parent_selector and type(parent_selector)==str:
          template = "document.querySelectorAll(`%s %s`).forEach(el => el.style[`%s`] = `%s`);"
          script = template % (parent_selector,selector,key,value)
        elif isinstance(parent_selector, WebElement):
          template = "arguments[0].querySelectorAll(`%s`).forEach(el => el.style[`%s`] = `%s`);"
          script = template % (selector,key,value)
        else:
          template = "document.querySelectorAll(`%s`).forEach(el => el.style[`%s`] = `%s`);"
          script = template % (selector,key,value)
        self.execute(script,parent_selector)
  
  # == module 5: size and positon == #
  def get_window_size(self):
    script = 'return {width:window.innerWidth,height:window.innerHeight};'
    return self.execute(script)

  def get_document_size(self):
    script = 'return {width:document.documentElement.offsetWidth,height:document.documentElement.offsetHeight};'
    return self.execute(script)

  def get_size(self,selector):
    script = 'var ele=document.querySelector(`%s`); return {width:ele.offsetWidth,height:ele.offsetHeight};' % selector
    return self.execute(script)
  
  # == module 6: alert and frame == #
  def alert(self,*args):
    '''
    @description : alert
    @param {tuple} args : one or more string 
    '''
    script = '''
    var alert_str = '';
    if(arguments.length>0){
      for(var i=0;i<arguments.length;i++){
        alert_str+=' '+arguments[i];
      }
    }
    return alert(alert_str);
    '''
    self.execute(script,*args)

  # == module 7: specific setting == #
  def highlight(self,selector):
    if not selector:
      return
    script = '''
      document.querySelector(`%s`).style.border="1px solid red";
    ''' % selector
    return self.execute(script)


  # == module 8: append elements == #
  def append(self,slot_selector,html):
    self.__insert_node(slot_selector,html,method="append")
  
  def prepend(self,slot_selector,html):
    self.__insert_node(slot_selector,html,method="prepend")

  def __insert_node(self,slot_selector,html,method="append"):
    '''
    use the createRange (not createElement) to avoid to add a middle div element
    '''
    script = '''
      var newElement = document.createRange().createContextualFragment(`%s`);
      document.querySelector(`%s`).%s(newElement);
    ''' % (html,slot_selector,method)
    return self.execute(script)

  # == module 9: exists == #
  def exists(self,selector):
    script = '''
    return document.querySelector(`%s`);
    ''' % selector
    return self.execute(script)




from blues_lib.material.MatHandler import MatHandler
from blues_lib.type.output.STDOut import STDOut
from blues_lib.util.BluesURL import BluesURL
from blues_lib.util.BluesAlgorithm import BluesAlgorithm

class Normalizer(MatHandler):

  def _calculate(self)->STDOut:
    valid_entities = []
    invalid_entities = []

    for entity in self._entities:
      
      self._add_system_fileds(entity)
      # set for detail only
      paras:list[dict] = entity.get('mat_paras')
      if paras:
        entity['mat_paras'] = self._get_format_paras(paras)
        # 如果没有缩略图，则从第一个段落中提取
        self._add_thumb(entity)
        
      # 如果没有thumb图片标识为错误 (目前用于成品判断，必须有图片)
      if not entity.get('mat_thumb'):
        entity['mat_stat'] = 'invalid'
        entity['mat_remark'] = 'no thumb image'
        invalid_entities.append(entity)
        continue
      
      valid_entities.append(entity)

    # insert invalid entities to db
    self._insert_trash(invalid_entities)

    message:str = f"valid {len(valid_entities)} ; invalid {len(invalid_entities)}" 
    return STDOut(200,message,valid_entities,invalid_entities)
  
  def _add_system_fileds(self,entity:dict):
    entity.setdefault('mat_chan',self._rule.get('mat_chan') or 'article')  # article gallery shortvideo qa
    if mat_url := entity.get('mat_url'):
      mat_site = BluesURL.get_main_domain(mat_url)
      mat_id = BluesAlgorithm.md5(mat_url)
      entity.setdefault('mat_site',mat_site)  # cn en
      entity.setdefault('mat_id',mat_id) 
      
  def _add_thumb(self,entity:dict):
    if not entity.get('mat_thumb'):
      for para in entity['mat_paras']:
        if para.get('type') == 'image':
          entity['mat_thumb'] = para['value']
          break

  def _get_format_paras(self,paras:list[dict])->list[dict]:
    
    # 如果是llm获取的就是结构化数组无需转换
    if paras[0].get('type'):
      return paras

    format_paras:list[dict] = []
    for row in paras:
      image = row.get('image')
      text = row.get('text')
      if image:
        format_paras.append({'type':'image','value':image})
      else:
        format_paras.append({'type':'text','value':text})
    return format_paras




from blues_lib.type.factory.Factory import Factory
from blues_lib.type.executor.Command import Command
from blues_lib.namespace.CommandName import CommandName

# flow
from blues_lib.command.flow.Engine import Engine as FlowEngine
from blues_lib.command.flow.Loop import Loop as FlowLoop

# llm
from blues_lib.command.llm.Engine import Engine as LLMEngine
from blues_lib.command.llm.Loop import Loop as LLMLoop

# crawler
from blues_lib.command.crawler.Engine import Engine as CrawlerEngine
from blues_lib.command.crawler.Loop import Loop as CrawlerLoop

# standard
from blues_lib.command.standard.Dummy import Dummy
from blues_lib.command.standard.Cleaner import Cleaner

# sql
from blues_lib.command.sql.Querier import Querier
from blues_lib.command.sql.Updater import Updater
from blues_lib.command.sql.Inserter import Inserter
from blues_lib.command.sql.Deleter import Deleter

# notifier
from blues_lib.command.notifier.Email import Email

# material
from blues_lib.command.material.Sinker import Sinker
from blues_lib.command.material.Deduplicator import Deduplicator
from blues_lib.command.material.Validator import Validator
from blues_lib.command.material.Normalizer import Normalizer
from blues_lib.command.material.Localizer import Localizer

class CommandFactory(Factory):

  _COMMANDS:dict[str,Command] = {
    # flow command
    FlowEngine.NAME:FlowEngine,
    FlowLoop.NAME:FlowLoop,

    # llm
    LLMEngine.NAME:LLMEngine,
    LLMLoop.NAME:LLMLoop,

    # crawler
    CrawlerEngine.NAME:CrawlerEngine,
    CrawlerLoop.NAME:CrawlerLoop,
    
    # standard
    Dummy.NAME:Dummy,
    Cleaner.NAME:Cleaner,
    
    # sql
    Querier.NAME:Querier,
    Updater.NAME:Updater,
    Inserter.NAME:Inserter,
    Deleter.NAME:Deleter,
    
    # notifier
    Email.NAME:Email,

    # material
    Sinker.NAME:Sinker,
    Deduplicator.NAME:Deduplicator,
    Validator.NAME:Validator,
    Normalizer.NAME:Normalizer,
    Localizer.NAME:Localizer,
    
  }
  
  _TASK_CONF_FIELDS:list[str] = [
    'task_id',
    'command',
  ]
  
  @classmethod
  def create(cls,task_def:dict,bizdata:dict,ti:any,context:dict|None=None)->Command | None:
    
    error:str = cls.check(task_def,bizdata,ti)
    if error:
      raise ValueError(f"Failed to create command - {error}")

    # overide
    command_value:str = task_def.get('command')
    command_name:CommandName|None = CommandName.from_value(command_value)
    if not command_name:
      error = f"The command '{command_value}' is not supported."
      raise ValueError(f"Failed to create command - {error}")

    executor:Command|None = cls._COMMANDS.get(command_name)
    if not executor:
      error = f"The command '{command_value}' is not supported."
      raise ValueError(f"Failed to create command - {error}")

    return executor(task_def,bizdata,ti,context)
  
  @classmethod
  def check(cls,task_def:dict,bizdata:dict,ti:any)->str:
    error:str = ''
    if not ti:
      error = f"{cls.__class__.__name__} The parameter 'ti' is missing."
    if not bizdata or not isinstance(bizdata,dict):
      error = f"{cls.__class__.__name__} The parameter 'bizdata' must be a dict."
    if not task_def or not isinstance(task_def,dict):
      error = f"{cls.__class__.__name__} The parameter 'task_def' must be a dict."

    for field in cls._TASK_CONF_FIELDS:
      if task_def.get(field) is None:
        error = f"{cls.__class__.__name__} The parameter 'task_def.{field}' is missing."   
        break
    return error




from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model
from blues_lib.namespace.CrawlerName import CrawlerName
from blues_lib.crawler.Crawler import Crawler
from blues_lib.crawler.base.BaseCrawlerFactory import BaseCrawlerFactory

class NestCrawler(Crawler):
  
  def _crawl_by_meta(self,model:Model,crawler_type:str='')->STDOut:
    # crawl : loop crawler will merge the entity to the output.data
    summary_conf:dict = model.config.get('summary',{})
    self._disable_quit(summary_conf)
    crawler_name:CrawlerName = self._get_crawler(summary_conf,crawler_type)
    crawler = BaseCrawlerFactory(model,self._browser).create(crawler_name)
    return crawler.execute()
  
  def _get_crawler(self,summary_conf:dict,crawler_type:str='')->CrawlerName:
    conf_crawler_type:str = summary_conf.get(CrawlerName.Field.TYPE.value)
    dft_crawler_type:str = CrawlerName.Engine.URL.value
    real_crawler_type:str = conf_crawler_type or crawler_type or dft_crawler_type
    return CrawlerName.Engine.from_value(real_crawler_type)
  
  def _disable_quit(self,summary_conf:dict):
    # 如何明确quit则保留，否则meta子crawler默认quit=False
    if summary_conf.get(CrawlerName.Field.QUIT.value)!= True:
      summary_conf[CrawlerName.Field.QUIT.value] = False



from abc import ABC,abstractmethod

class ChromeArgs(ABC):

  def __init__(self,args=None):
    '''
    @param {dict} args : the input experimental args dict, such as {'debugaddr':'127.0.0.1:9222'}
      - the dict's key is not the real option attr
    ''' 
    self._input_args = args 

  @abstractmethod 
  def get():
    pass

  @abstractmethod 
  def get_from_input():
    pass



from blues_lib.command.NodeCommand import NodeCommand
from blues_lib.namespace.CommandName import CommandName
from blues_lib.type.output.STDOut import STDOut
from blues_lib.material.privatizer.Localizer import Localizer as LocalizerHandler

class Localizer(NodeCommand):
  NAME = CommandName.Material.LOCALIZER

  def _invoke(self)->STDOut:
    rule:dict = self._summary.get('rule',{})
    entities:list[dict] = self._summary.get('entities',[])
    if not entities:
      raise ValueError(f'{self.NAME} : no input entities')

    request = {
      'rule':rule,
      'entities':entities, # must be a list
    } 
    handler = LocalizerHandler(request)
    return handler.handle()




import os
import json,csv
from blues_lib.util.BluesFiler import BluesFiler

class FileWriter:

  @classmethod
  def write(cls,file_path,text,mode='w')->str:
    '''
    @description : write text to file
    @param {str} file_path : file's path
    @param {str} text : content
    @param {str} mode : write mode
      - 'w' : clear the history content
      - 'a' : append text
    @returns {str} : the writed file path
    '''
    dir_path = os.path.dirname(file_path)
    if dir_path!='.':
      BluesFiler.makedirs(dir_path)

    try:
      with open(file_path,mode,encoding='utf-8') as file:
        file.write(text)
      return file_path
    except Exception as e:
      print(f'==>write error: {e}')
      return ''

  @classmethod
  def write_after(cls,file_path,text)->str:
    return cls.write(file_path,text,'a')

  @classmethod
  def write_json(cls,file_path,data,indent=2)->str:
    dir_path = os.path.dirname(file_path)
    if dir_path!='.':
      BluesFiler.makedirs(dir_path)

    try:
      with open(file_path, 'w', encoding='utf-8') as file:
        # must set ensure_ascii as False for Chinese character
        json.dump(data,file,indent=indent,ensure_ascii=False)
        return file_path
    except Exception as e:
      print(f'==>write_json error: {e}')
      return ''

  @classmethod
  def write_csv(cls,file_path,rows,header=None,mode='w')->str:
    '''
    @description : write to csv file
    @param {str} file_path : the csv file's path
    @param {list<list>|tuple<tuple>} rows : the data rows
    @param {list|tuple} header
    @param {str} mode : 'a' - append ; 'w' - cover
    '''
    dir_path = os.path.dirname(file_path)
    if dir_path!='.':
      BluesFiler.makedirs(dir_path)

    try:
      with open(file_path, mode, encoding='utf-8',newline="") as file:
        writer = csv.writer(file) 
        if mode=='w' and header:
          writer.writerow(header)
        if rows:
          for row in rows:
            writer.writerow(row)
        return file_path
    except Exception as e:
      print(f'==>write_csv error: {e}')
      return ''



from blues_lib.type.output.SQLSTDOut import SQLSTDOut
from blues_lib.dao.sql.BluesMySQL import BluesMySQL
from blues_lib.dao.sql.BluesSQLConvertor import BluesSQLConvertor

class BluesSQLIO():

  def __init__(self,account:dict=None):
    '''
    @param {dict} account 
    {
      'host':'localhost',
      'user':'root',
      'password':'',
      'database':'infocollection',
    }
    '''
    self.sql_executor = BluesMySQL.get_instance(account)

  def get(self,table:str,fields="*",conditions=None,orders=None,pagination=None)->SQLSTDOut:
    '''
    @description 查询用户提交数据
    @param {dict[]} conditions : one or more conditions
      [
        {'operator':'and','field':'name','comparator':'=','value':'blues','value_type':'str'},
        {'operator':'and','field':'name','comparator':'=','value':'blues'}
      ] 
    @param {dict[]} : orders
      [{'field':'id','sort':'asc'},{'field':'name','sort':'desc'}] 
    @param {dict} pagination : page info
      {'no':1,'size':10}
    @returns {SQLSTDOut}
    '''
    condition_sql = BluesSQLConvertor.get_condition_sql(conditions)
    order_sql = BluesSQLConvertor.get_order_sql(orders)
    limit_sql = BluesSQLConvertor.get_limit_sql(pagination)
    if isinstance(fields,str):
      field_sql = fields 
    elif isinstance(fields,list) or isinstance(fields,tuple): 
      field_sql = ','.join(fields)
    else:
      field_sql = '*'

    sql = 'select %s from %s %s %s %s' % (field_sql,table,condition_sql,order_sql,limit_sql)  
    return self.sql_executor.get(sql) 

  def insert(self,table:str,entities:list[dict])->SQLSTDOut:
    '''
    @description : insert by sql
    @param {dict | dict[]} entities : the entity dict's key is the real table field
      [{'name':'blues','age':18}]
    @returns {SQLSTDOut} 
    '''
    insert_sql = BluesSQLConvertor.get_insert_sql(entities)
    sql = 'insert into %s %s' % (table,insert_sql)
    return self.sql_executor.post(sql) 
 
  def post(self,table:str,fields:list,values:list[list])->SQLSTDOut:
    '''
    @description : insert by template sql
    @param {list|tuple} fields : the fields will be updated
      ['name','age']
    @param {list | list[]} values : one or multi row data values
      [['post01',1],['post02',2]]
    @returns {SQLResult} 
    '''
    insert_sql = BluesSQLConvertor.get_insert_template_sql(fields)
    sql = 'insert into %s %s' % (table,insert_sql)
    return self.sql_executor.post(sql,values) 
  
  def put(self,table:str,fields:list,values:list,conditions:list[dict])->SQLSTDOut:
    '''
    @description : update row
    @param {list|tuple} fields : the fields will be updated
      ['name','age']
    @param {list|tuple} values : the values will be writed
      ['blues',18]
    @param {dict[]} conditions : one or more conditions
      [
        {'operator':'and','field':'name','comparator':'=','value':'blues'},
        {'operator':'and','field':'name','comparator':'=','value':'blues'}
      ] 
    @returns {SQLResult} 
    '''
    update_sql = BluesSQLConvertor.get_update_template_sql(fields,conditions)
    sql = 'update %s %s' % (table,update_sql)
    return self.sql_executor.put(sql,values) 

  def update(self,table:str,entity:dict,conditions:list[dict])->SQLSTDOut:
    '''
    @description : update row
    @param {dict} entity : the entity dict's key is the real table field
      {'name':'blues','age':18}
    @param {dict[]} conditions : one or more conditions
      [
        {'operator':'and','field':'name','comparator':'=','value':'blues'},
        {'operator':'and','field':'name','comparator':'=','value':'blues'}
      ] 
    @returns {SQLResult} 
    '''
    update_sql = BluesSQLConvertor.get_update_sql(entity,conditions)
    sql = 'update %s %s' % (table,update_sql)
    return self.sql_executor.put(sql) 
  
  def delete(self,table:str,conditions:list[dict])->SQLSTDOut:
    '''
    @description : delete rows by conditon 
    @param {str} table 
    @param {dict[]} conditions : one or more conditions
      [
        {'operator':'and','field':'name','comparator':'=','value':'blues'},
        {'operator':'and','field':'name','comparator':'=','value':'blues'}
      ] 
    @returns {SQLResult}
    '''
    condition_sql = BluesSQLConvertor.get_condition_sql(conditions)
    sql = 'delete from %s %s' % (table,condition_sql)
    return self.sql_executor.delete(sql)   



import re
from blues_lib.util.BluesDateTime import BluesDateTime
from blues_lib.hocon.replacer.BaseReplacer import BaseReplacer

class FunctionReplacer(BaseReplacer):

  '''
  Calculated variable replacer that replaces variable placeholders with values from get_vars method.
  Uses the same format as BaseVarReplacer but gets values from a dictionary returned by get_vars.
  "%{function timestamp}" 内置方法
  "%{lambda x: x.upper()}" 自定义函数 : 暂未实现
  '''
  FUNCTION_PATTERN = r"^%\{function\s+([a-zA-Z0-9_]+)\}$"
  FUNCS = {
    "timestamp": lambda: BluesDateTime.get_timestamp(),
    "now": lambda: BluesDateTime.get_now(),
  }

  def _replace_node_var(self, value:str)->str:
    """替换单个字符串中的function占位符（要求完全匹配）"""
    # 检查是否为完整的 function 变量
    match = re.match(self.FUNCTION_PATTERN, value)
    if not match:
      return value

    func_name = match.group(1)  
    if func_name in self.FUNCS:
      return self.FUNCS[func_name]()
    else:
      return value 



from blues_lib.behavior.Trigger import Trigger

class Remove(Trigger):

  def _trigger(self)->any:
    kwargs = self._get_kwargs(['loc_or_elem'])
    return self._browser.element.popup.remove(**kwargs)


import sys,os,re,time
from abc import ABC,abstractmethod
from functools import wraps
from selenium.webdriver.remote.webelement import WebElement

from blues_lib.util.BluesConsole import BluesConsole

class WaiterDeco(ABC):

  caller_class = 'EC'

  def __init__(self,caller_method,level=3):
    '''
    Create a special decorator for class waiter.EC 
    Parameter:
      caller_method {str} : the wrappered func's name
      level {int} : the log level
        - 3 error : only error info
        - 2 info : error and info
        - 1 debug : all info
    '''
    self.caller_method = caller_method
    self.level = level 

    # func's args index
    self.arg_cs_index = None 
    self.arg_pcs_index = None
    self.arg_value_index = None

    # rewrite the index
    self.set_arg_index()
  
  @abstractmethod
  def set_arg_index(self):
    '''
    Template method
    Get the caller func's timeout parameter's index
    '''
    pass

  def wrapper(self,func,*args,**kwargs):
    # execute the wrappered func
    start_time = time.time()
    outcome = func(*args,**kwargs)
    end_time = time.time()

    elapsed_time = round(end_time - start_time,2)
    msgs = self.get_finish_msg(outcome,args,elapsed_time)

    # print success msg
    if outcome and self.level<=2 :
      BluesConsole.success(msgs[0])

    # print failure msg
    if not outcome and self.level<=3:
      BluesConsole.error(msgs[1])
    
    # must return the wrappered func's value
    return outcome

  def get_finish_msg(self,outcome,args,elapsed_time):
    info = self.get_info(args)
    info.update({ 
      'duration':elapsed_time, 
      'type':type(outcome).__name__,
    })

    values = (self.caller_class,self.caller_method,info)
    s_msg = '[%s.%s] Success %s' % values
    f_msg = '[%s.%s] Failure %s' % values

    return [s_msg,f_msg]
   
  def get_info(self,args):
    cs = self.get_cs(args)
    pcs = self.get_pcs(args)
    value = self.get_value(args)
    info = {}
    if cs:
      info['cs'] = cs
    if pcs:
      info['pcs'] = pcs
    if value:
      info['value'] = str(value)
    return info

  def get_cs(self,args):
    index = self.arg_cs_index
    return self.get_cs_title(args,index)

  def get_pcs(self,args):
    index = self.arg_pcs_index
    return self.get_cs_title(args,index)

  def get_cs_title(self,args,index):
    if index and len(args)>index:
      CS_WE = args[index]
      if type(CS_WE) == WebElement:
        return 'WebElement %s' % CS_WE.tag_name
      else:
        return CS_WE
    else:
      return None

  def get_value(self,args):
    index = self.arg_value_index
    if index and len(args)>index:
      return args[index]
    else:
      return None



import sys,os,re
from typing import Any

from blues_lib.behavior.Bean import Bean

class Choice(Bean):

  def _set(self)->Any:
    kwargs = self._get_kwargs(['loc_or_elem','parent_loc_or_elem','timeout'])
    is_select = self._config.get('value',True)
    if self._to_be_clickable():
      if is_select:
        return self._browser.element.choice.select(**kwargs) 
      else:
        return self._browser.element.choice.deselect(**kwargs)



from blues_lib.sele.waiter.Querier import Querier  
from blues_lib.sele.element.Info import Info  
from blues_lib.util.BluesFiler import BluesFiler
from blues_lib.util.BluesDateTime import BluesDateTime

class File():

  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver,5) 
    self.__info = Info(driver) 

  def write(self,loc_or_elem,value,wait_time=3,parent_loc_or_elem=None,timeout=5):
    '''
    Add one or multiple files to the file input
    If there are multiple files, the upload mode is controlled based on whether multiple file upload is supported
    '''

    files = value if type(value) == list else [value]
    # Supports uploading multiple images at a time
    exist_files = BluesFiler.filter_exists(files)
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not exist_files or not web_element:
      return

    self._set_element_visible(web_element)

    is_multiple = self.__info.get_attr(web_element,'multiple')
    if is_multiple:
      # must join the file paths by \n
      file_lines = '\n'.join(exist_files)
      web_element.send_keys(file_lines)
      BluesDateTime.count_down({'duration':wait_time,'title':'Wait image upload...'})
    else:
      for exist_file in exist_files:
        web_element.send_keys(exist_file)
        BluesDateTime.count_down({'duration':wait_time,'title':'Wait image upload...'})
        
  def _set_element_visible(self,web_element):
    """
    确保文件输入框可见且具备可交互尺寸
    - 若元素隐藏(display: none)，设置为inline
    - 若元素尺寸为0(width/height=0)，设置为auto
    """
    driver = self.__driver
    
    # 获取当前样式属性
    current_display:bool = web_element.is_displayed()
    # 获取元素尺寸（rect返回字典包含width/height）
    element_rect:dict = web_element.rect
    current_width = element_rect['width']
    current_height = element_rect['height']
    
    # 调整display属性（如果隐藏）
    if not current_display:
      driver.execute_script("arguments[0].style.display = 'inline';", web_element)
    
    # 调整尺寸（如果为0或接近0）
    min_size = 1  # 最小有效尺寸阈值（像素）
    if current_width <= min_size:
      driver.execute_script("arguments[0].style.width = 'auto';", web_element)
    if current_height <= min_size:
      driver.execute_script("arguments[0].style.height = 'auto';", web_element)



from blues_lib.behavior.Trigger import Trigger
from blues_lib.model.Model import Model
from blues_lib.behavior.BhvExecutor import BhvExecutor
from blues_lib.types.common import LightBhvList,LightBhvItem

class LightChain(Trigger):
  # light weight chain, only support core attributes
  def _trigger(self)->int:
    light_bhvs:LightBhvList = self._config.get('bhv_chain')
    if not light_bhvs:
      return 0
    
    index = 0
    for light_bhv in light_bhvs:
      self._trigger_bhv(light_bhv)
      if index<len(light_bhvs)-1:
        self._wait_interval(index)
      index += 1
    return index
  
  def _trigger_bhv(self,light_bhv:LightBhvItem)->int:
    bhv_list = light_bhv + [None] * (3 - len(light_bhv))
    meta = {
      '_kind':bhv_list[0],
      'loc_or_elem':bhv_list[1],
      'value':bhv_list[2],
    }
    self._padd_default_attr(meta)
    model = Model(meta)
    return BhvExecutor(model,self._browser).execute()
  
  def _padd_default_attr(self,meta:dict):
    if meta.get('_kind') == 'write_text':
      meta['clearable'] = True


import os
import platform

class Process:
  # 提供跨系统进程操作的方法
  
  @classmethod
  def kill_chrome(cls):
    '''
    Stop all chrome process
    @returns {int} : 0-success >0- failure
    '''
    return cls.kill('chrome')
  
  @classmethod
  def kill(cls, task_name:str):
    '''
    Stop a app's all process
    @param {str} task_name : the app's process name
    @returns {int} : 0-success >0- failure
    '''
    system = platform.system().lower()
    
    try:
      if 'windows' in system:
        # Windows system
        return os.system(f'taskkill /F /IM {task_name}.exe')
      elif 'darwin' in system:
        # macOS system
        return os.system(f'pkill -f {task_name}')
      elif 'linux' in system:
        # Linux system
        return os.system(f'pkill -f {task_name}')
      else:
        # Unsupported system
        print(f"Unsupported operating system: {system}")
        return 1
    except Exception as e:
      print(f"Error occurred while killing process: {e}")
      return 1



import hashlib
class BluesAlgorithm():

  @classmethod
  def md5(cls,text):
    # 单向不可逆，作为数据存储密码，如果忘了只能重置
    md_5 = hashlib.md5()
    # update入参必须时 bytes类型
    md_5.update(text.encode())
    return md_5.hexdigest()
  
  @classmethod
  def sha1(cls,text):
    sha_1 = hashlib.sha1()
    sha_1.update(text.encode())
    return sha_1.hexdigest()


import re
from typing import Any
from blues_lib.hocon.replacer.BaseReplacer import BaseReplacer

class VariableReplacer(BaseReplacer):
 
  IS_VARIABLES_REQUIRED = True

  '''
  replace the variable replace with the given values
  - single variable "{{var}}"
  - interpolate variable "name is {{name}}, age is {{age}}"
  '''  
  VAR_PATTERN = r"\{\{([a-zA-Z0-9_:\-.]+)\}\}"
  
  def _replace_node_var(self, value:str)->str:
    match = re.fullmatch(self.VAR_PATTERN, value)
    if match:
      # single variable replace, use the original value or None
      expr = match.group(1) 
      key, default = self._get_key_default(expr)
      return self._variables.get(key, default)
    else:
      return self._interpolate(value)
    
  def _interpolate(self, value:str)->str:
    # 处理包含多个占位符或其他文本的字符串
    def replace_match(match):
      expr = match.group(1)
      # 处理带默认值的情况: ${key:-default}
      key, default = self._get_key_default(expr)
      return str(self._variables.get(key, default))
     
    return re.sub(self.VAR_PATTERN, replace_match, value)
  
  def _get_key_default(self, expr:str,as_string:bool=False)->tuple:
    # 处理带默认值的情况: {{key:-default}}
    if ':-' in expr:
      key, default = expr.split(':-', 1)
      value = default if as_string else self._get_raw_value(default)
      return key, value
    return expr, None

  def _get_raw_value(self, value_str: str) -> Any|None:
    """
    将字符串转换为原始类型
    now ,don't support list and dict
    """
    # 处理布尔值
    if value_str.lower() == 'true':
      return True
    if value_str.lower() == 'false':
      return False
    
    # 处理None
    if value_str.lower() == 'none':
      return None
    
    # 处理数字
    try:
      num = int(value_str)
      return num
    except ValueError:
      try:
        num = float(value_str)
        return num
      except ValueError:
        pass
    # 默认返回字符串
    return value_str



from blues_lib.hook.command.CommandProc import CommandProc
class Blocker(CommandProc):
  
  def execute(self)->None:
    '''
    @description: block the flow
    @return: None
    '''
    should_block:bool = self._proc_conf.get('should_block',False)
    if should_block:
      self._ti.xcom_push('should_block',True)



from functools import wraps
from .InfoDeco import InfoDeco

class SelectDeco(InfoDeco):

  caller_class = 'Select'

  def __call__(self,func):
    @wraps(func) 
    def wrapper(*args,**kwargs):
      return self.wrapper(func,*args,**kwargs)
    return wrapper




import time

from blues_lib.behavior.Trigger import Trigger
from blues_lib.util.Clipboard import Clipboard

class Keyboard(Trigger):

  def _trigger(self)->str:
    key:str = self._config.get('key')
    value:str = self._config.get('value')
    kwargs = self._get_kwargs(['loc_or_elem','parent_loc_or_elem'])
    
    actions:list[str] = ['select','focus','copy','cut','clear','enter','f12']

    if key == 'paste':
      self._paste(kwargs,value)
    elif key == 'paste_after':
      self._paste(kwargs,value,True)
    elif key == 'esc':
      self._browser.action.keyboard.esc()
    elif key in actions:
      func = getattr(self._browser.action.keyboard,key)
      func(**kwargs)

  def _paste(self,kwargs:dict,value:str,after:bool=False):
    if value:
      Clipboard.copy(value) # write to clipboard
    time.sleep(0.2)
    # write to the input
    if after:
      self._browser.action.keyboard.paste_after(**kwargs)
    else:
      self._browser.action.keyboard.paste(**kwargs)




import sys,os,re
from .Finder import Finder   
from .deco.StateDeco import StateDeco

# 提供元素选择功能
class State():

  def __init__(self,driver):
    self.__driver = driver
    self.__finder = Finder(driver)

  # === part 4: get element's state, don't wiat === #
  @StateDeco('is_presence')
  def is_presence(self,loc_or_elem,parent_loc_or_elem=None):
    '''
    Detemines if the element is in the document
    It can be hidden
    Parameter:
      loc_or_elem,parent_loc_or_elem {str|WebElement} : css selector or WebElement
    Returns:
      {bool}
    '''
    return True if self.__finder.find(loc_or_elem,parent_loc_or_elem) else False

  @StateDeco('is_displayed')
  def is_displayed(self,loc_or_elem,parent_loc_or_elem=None):
    '''
    If the connected Element is displayed or not displayed on a webpage
    Parameter:
      loc_or_elem,parent_loc_or_elem {str|WebElement} : css selector or WebElement
    Returns:
      {bool}
    '''
    web_element = self.__finder.find(loc_or_elem,parent_loc_or_elem)
    if not web_element:
      return False
    return web_element.is_displayed()

  @StateDeco('is_enabled')
  def is_enabled(self,loc_or_elem,parent_loc_or_elem=None):
    '''
    If the connected Element is enabled or disabled on a webpage
    Parameter:
      loc_or_elem,parent_loc_or_elem {str|WebElement} : css selector or WebElement
    Returns:
      {bool}
    '''
    web_element = self.__finder.find(loc_or_elem,parent_loc_or_elem)
    if not web_element:
      return False
    return web_element.is_enabled()

  @StateDeco('is_selected')
  def is_selected(self,loc_or_elem,parent_loc_or_elem=None):
    '''
    Detemines if the element is selected or not, widely used on:
      1. check box
      2. radio box
      3. input element
      4. option element
    Parameter:
      loc_or_elem,parent_loc_or_elem {str|WebElement} : css selector or WebElement
    Returns:
      {bool}
    '''
    web_element = self.__finder.find(loc_or_elem,parent_loc_or_elem)
    if not web_element:
      return False
    return web_element.is_selected()



import random,time
from blues_lib.behavior.write.Write import Write

class ResetChar(Write):

  _set_keys = ['loc_or_elem','value','parent_loc_or_elem','timeout']

  def _set(self)->any:
    # clear the input and input a new char
    value:str = str(self._config.get('value'))  or ''
    scope:list[int|float] = self._get_interval_scope([0.5,0.5])
    return self._input_char(value,scope)

  def _input_char(self,value:str,scope:list[int|float])->any:
    # 这里必须重新选择元素，应对豆包登录，输入一个字符后元素重新绘制问题
    for char in value:
      config = {
        'value':char
      }
      kwargs = self._get_kwargs(self._set_keys,config)

      # must wait some seconds, to wait the element replaced
      interval = random.uniform(*scope)
      time.sleep(interval)

      # select the element again
      if self._to_be_visible():
        # the element always only has one char
        self._browser.element.input.write(**kwargs)


from blues_lib.command.LoopCommand import LoopCommand
from blues_lib.namespace.CommandName import CommandName
from blues_lib.type.output.STDOut import STDOut
from blues_lib.model.Model import Model
from blues_lib.command.crawler.Invoker import Invoker

class Loop(LoopCommand):

  NAME = CommandName.Crawler.LOOP

  def _run_once_cal(self,model:Model)->STDOut:
    return Invoker.invoke(model,self._context)
  


from blues_lib.hook.command.processor.material.MatProc import MatProc
from blues_lib.material.normalizer.Normalizer import Normalizer
from blues_lib.type.output.STDOut import STDOut

class MatNormalizer(MatProc):
  
  def _calculate(self,request:dict)->STDOut:
    handler = Normalizer(request)
    return handler.resolve()


from abc import ABC,abstractmethod

class CleanerHandler(ABC):

  def __init__(self):
    '''
    The abstract class of handlers 
    '''
    self._next_handler = None
  
  def set_next(self,handler):
    '''
    Set the next handler
    Parameter:
      handler {Handler} : the next handler
    Returns 
      {Handler} 
    '''
    self._next_handler = handler
    return handler

  def handle(self,request):
    '''
    The full cleanup handlers chain
    Args:
      request {dict} : 
        - db {dict}
        - file {dict}
    Returns 
      {dict} : this last handler's handled response
    '''
    response =self.resolve(request)
    # It's a pipeline, don't stop, go throght all handlers
    if self._next_handler:
      return self._next_handler.handle(request)
    else:
      return response

  @abstractmethod
  def resolve(self,data):
    '''
    This method will be implemented by subclasses
    '''
    pass


import re
from blues_lib.util.ResourceReader import ResourceReader
from blues_lib.util.DotFileReader import DotFileReader
from blues_lib.hocon.replacer.BaseReplacer import BaseReplacer

class IncludeReplacer(BaseReplacer):

  '''
  include 文件内容替换
  "%{include schema.except.stdout.200_list}" 内置json schema文件
  "%{include template.metadata.crawler.html}" 内置hocon template文件
  '''  
  INCLUDE_PATTERN = r"^%\{include\s+(schema|template)\.([a-zA-Z0-9_.-]+)\}$"
  
  def _replace_node_var(self, value:str)->str:
    """替换单个字符串中的占位符（要求完全匹配）"""
    # 检查是否为完整的 include 变量
    match = re.match(self.INCLUDE_PATTERN, value)
    if not match:
      return value

    root_name = match.group(1)  # 提取根目录部分
    dot_path = match.group(2)  # 提取文件dot路径 
    package = f'blues_lib.{root_name}' # pad the package name
    return self._read_template(package,dot_path)
    
  def _read_template(self,package:str,dot_path:str)->dict:
    extension:str = 'conf' # must be conf file
    slash_path:str = DotFileReader.get_slash_path(dot_path,extension)
    return ResourceReader.read_hocon(package, slash_path)



import sys,os,re

from blues_lib.sele.element.Finder import Finder 
from blues_lib.sele.element.State import State 
from blues_lib.sele.interactor.Frame import Frame 
from blues_lib.sele.script.javascript.JavaScript import JavaScript 

class Popup():
  '''
  Handle automatic pop-up AD boxes
  '''
  def __init__(self,driver):
    self.__driver = driver
    self.__finder = Finder(driver)
    self.__state = State(driver)
    self.__frame = Frame(driver)
    self.__javascript = JavaScript(driver)

  def remove(self,loc_or_elem):
    '''
    Remove the popup elements
    Parameter:
      loc_or_elem {list<str>} : target_CSs selector list
    Returns:
      {int} : removed count
    '''
    count = 0
    if not loc_or_elem:
      return count
    
    loc_or_elems = loc_or_elem if type(loc_or_elem) == list else [loc_or_elem]

    for CS_WE in loc_or_elems:
      count += self.__javascript.remove(CS_WE)

    return count

  def close(self,loc_or_elem,parent_loc_or_elem=None,frame_loc_or_elem=None):
    '''
    Close one or multi popups
    Parameter:
      off_locators { list<[close_CS_WE,frame_loc_or_elem]> | list<close_CS_WE>} 
    Returns:
      {int} : the closed count
    '''
    count = 0
    if not loc_or_elem:
      return count

    loc_or_elems = loc_or_elem if type(loc_or_elem)==list else [loc_or_elem]
    
    for CS_WE in loc_or_elems:
      count += self.__close(CS_WE,parent_loc_or_elem,frame_loc_or_elem)

    return count

  def __close(self,close_CS_WE,parent_loc_or_elem=None,frame_loc_or_elem=None):
    count = 0
    # switch to frame
    if frame_loc_or_elem:
      self.__frame.switch_to(frame_loc_or_elem)      

    web_elements = self.__finder.find_all(close_CS_WE,parent_loc_or_elem)
    if not web_elements:
      # switch back
      if frame_loc_or_elem:
        self.__frame.switch_to_default()      

      return count
    
    for web_element in web_elements:
      if self.__state.is_displayed(web_element):
        count+=1
        self.__mouse.click(web_element)
    
    # switch back
    if frame_loc_or_elem:
      self.__frame.switch_to_default()      

    return count





from blues_lib.sele.waiter.Querier import Querier  
from selenium.webdriver import ActionChains
from selenium.webdriver.common.actions.action_builder import ActionBuilder

class Mouse():
  '''
  If the element is outside the viewable window, 
  The element will automatically roll into the window, 
  With the bottom of the element flush with the bottom of the window
  '''
  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver)
    self.__chains = ActionChains(driver)
    self.__builder = ActionBuilder(driver)

  # == module 1 : click event == #
  # add a  hover deco
  # other method: hover().pause(1).click().perform()
  def click(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Click and release
    Parameter:
      loc_or_elem {str | WebElement}
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return
    # 更可靠滚动到元素可见位置，暂不使用chains实现
    web_element.click()
  
  def right_click(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Right click and release
    Parameter:
      loc_or_elem {str | WebElement}
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return
    self.__chains.context_click(web_element).perform()

  def double_click(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Double left click
    Parameter:
      loc_or_elem {str | WebElement}
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return
    self.__chains.double_click(web_element).perform()

  # == module 2 : hold and release == #
  def hold(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Click and hold
    Parameter:
      loc_or_elem {str | WebElement}
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return
    self.__chains.click_and_hold(web_element).perform()

  def release(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Release the hold
    Parameter:
      loc_or_elem {str | WebElement}
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    self.__chains.release(web_element).perform()

  # == module 3 : move into a element == #
  def move_in(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Moves the mouse to the in-view center point of the element
    If the element outside in the viewport, it will move into the viewport automatically
    Automaic moving : the element's bottom border align to the window's bottom border
    Parameter:
      loc_or_elem {str | WebElement}
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return
    self.__chains.move_to_element(web_element).perform()

  def move_in_offset(self,loc_or_elem,offset_x=0,offset_y=0,parent_loc_or_elem=None,timeout=5):
    '''
    Offset form the element
    This method moves the mouse to the in-view center point of the element, 
    Then moves by the provided offset.
    '''
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return
    self.__chains.move_to_element_with_offset(web_element,offset_x,offset_y).perform()

  # == module 4 : move by the viewport == #
  def move_to(self,x=0,y=0):
    '''
    Offset from Viewport
    The viewport's top left point to the element's center point (not the element's top left point)
    Test with hold
    '''
    self.__builder.pointer_action.move_to_location(x,y)
    self.__builder.perform()

  def move_to_offset(self,offset_x=0,offset_y=0):
    '''
    Offset from Viewport
    This method moves the mouse from its current position by the offset provided by the user. 
    Test with hold
    '''
    self.__chains.move_by_offset(offset_x,offset_y).perform()




from blues_lib.util.BluesDateTime import BluesDateTime
from blues_lib.behavior.Trigger import Trigger

class Wait(Trigger):

  def _trigger(self)->any:
    kwargs = self._get_kwargs(['duration','title'])
    return BluesDateTime.count_down(kwargs)


from blues_lib.util.NestedDataMapping import NestedDataMapping
from blues_lib.command.io.IOExcept import IOExcept

class InputMapper:

  def __init__(self,task_def:dict,bizdata:dict,ti:any) -> None:
    self._task_def = task_def
    self._bizdata = bizdata
    self._ti = ti

  def handle(self):
    # Base on input mapping, add upsteam's output to the bizdata by xcom

    IOExcept.validate_task_input(self._task_def)
    
    input_defs:list[dict] = self._task_def.get('input')
    if not input_defs:
      return

    for input_def in input_defs:
      self._map(input_def)
      
  def _map(self,input_def:dict):
    source:str = input_def.get('source')
    target:str = input_def.get('target')
    method:str = input_def.get('method') or 'assign' # cover the node
    before_map:str = input_def.get('before_map')
    
    if not source or not target:
      return
    source_data,source_attr_chain = self._get_source_data(source)
    if not source_data:
      return
    
    target_data,target_attr_chain = self._get_target_data(target)
    if not target_data:
      return
    has_mapped = NestedDataMapping.map(source_data,source_attr_chain,target_data,target_attr_chain,method,before_map) 
    if not has_mapped:
      raise Exception(f'Failed to map source {source} to target {target}')
      
  def _get_target_data(self,target:str)->tuple[any,str]:
    target_slices:list[str] = target.split('/')
    if len(target_slices) < 1:
      return None,''
    
    target_root:str = target_slices[0]
    attr_chain:str = target_slices[1] if len(target_slices) > 1 else ''
    
    if target_root !='bizdata':
      return None,attr_chain

    bizdata = self._bizdata
    return bizdata,attr_chain
    
  def _get_source_data(self,source:str)->tuple[any,str]:
    source_slices:list[str] = source.split('/')
    if len(source_slices) < 2:
      return None,''
    
    task_id:str = source_slices[0]
    key:str = source_slices[1]
    attr_chain:str = source_slices[2] if len(source_slices) > 2 else ''
    source_root_data:any = self._ti.xcom_pull(task_ids=task_id,key=key)
    if not source_root_data:
      return None,attr_chain 
    
    return source_root_data,attr_chain



import sys,os,re

from blues_lib.sele.waiter.Querier import Querier 
from blues_lib.sele.action.Mouse import Mouse 

class Choice():
  
  def __init__(self,driver):
    self.__driver = driver
    self.__querier = Querier(driver,5)
    self.__mouse = Mouse(driver)

  def select(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Support multi input parameters as selectors
    Parameter:
      loc_or_elem {str|WebElemnt|list<str>|list<WebElement>} : one or a list of target element
      parent_loc_or_elem {str|WebElemnt} : the choicebox's parent element
    Returns:
      {int} : the selectd count
    '''
    count = 0
    if not loc_or_elem:
      return count

    loc_or_elems = loc_or_elem if type(loc_or_elem) == list else [loc_or_elem]
    for loc_or_elem in loc_or_elems:
      count += self.__toggle(loc_or_elem,True,parent_loc_or_elem,timeout)
    return count

  def deselect(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    '''
    Support multi input parameters as selectors
    Parameter:
      loc_or_elem {str|WebElemnt|list<str>|list<WebElement>} : one or a list of target element
    Returns:
      {int} : the selectd count
    '''
    count = 0
    if not loc_or_elem:
      return count
    
    loc_or_elems = loc_or_elem if type(loc_or_elem) == list else [loc_or_elem]

    for loc_or_elem in loc_or_elems:
      count += self.__toggle(loc_or_elem,False,parent_loc_or_elem,timeout)
    return count

  def __toggle(self,loc_or_elem,checked=True,parent_loc_or_elem=None,timeout=5):
    '''
    Select the choice boxes by selectors
    Parameter
      loc_or_elem {str|WebElement} : boxes css selectors or WebElement
      - Maybe one element: 'iput[value=car]'
      - Maybe multi elements: 'input[value=car],input[value=boat]'
    Returns:
      {int} : selectd count
    '''
    count = 0
    web_elements = self.__querier.query_all(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_elements:
      return count
    
    for web_element in web_elements:
      # select mode
      if checked and web_element.is_selected():
        continue
      # deselect mode
      if not checked and not web_element.is_selected():
        continue
      count+=1
      # use the action to roll in the element to viewport automatically
      self.__mouse.click(web_element)
    return count

  def is_selected(self,loc_or_elem,parent_loc_or_elem=None,timeout=5):
    web_element = self.__querier.query(loc_or_elem,parent_loc_or_elem,timeout)
    if not web_element:
      return False
    return web_element.is_selected()



import sys,os,re

from blues_lib.util.BluesURL import BluesURL  
from blues_lib.util.BluesConsole import BluesConsole  

# 提供Cookie相关功能
class Cookie():
  '''
  WebDriver API provides a way to interact with cookies
  Cookie doc: https://www.selenium.dev/documentation/webdriver/interactions/cookies/
  '''
 
  def __init__(self,driver):
    self.__driver = driver

  def __get_default_entity(self):
    current_domain = BluesURL.get_main_domain(self.__driver.current_url)
    return {
      'name':'',
      'value':'',
      'domain':'.%s' % current_domain, 
      'expires': '',
      'path': '/',
      'httpOnly': False,
      'hostOnly': False,
      'secure': False,
    }

  def get(self,name=''):
    '''
    Get a named cookie or all cookies
    Parameter:
      name {str} : cookie name
    Returns: cookie entity or entities
       - {'domain': 'mp.163.com', 'httpOnly': True, 'name': 'NTESwebSI', 'path': '/', 'sameSite': 'Lax', 'secure': False, 'value': '2A7C7F8FCD65F7D74650E13D349C60CC'}
       - [{'domain': 'mp.163.com', 'httpOnly': True, 'name': 'NTESwebSI', 'path': '/', 'sameSite': 'Lax', 'secure': False, 'value': '2A7C7F8FCD65F7D74650E13D349C60CC'}]
    '''
    if name:
      return self.__driver.get_cookie(name)
    else:
      return self.__driver.get_cookies()
      
  def add(self,entities):
    '''
    Add cookies to the current browsing context
    The list of accepted JSON key values : https://www.w3.org/TR/webdriver1/#cookies
    Parameter:
      entities {list<dict> | dict} : one or multi cookie entities, it's a format cookie dict
      - {'name':'sessionid', 'value':'1715957318'}
    Returns:
      {int} : The number of successful additions
    '''
    
    count = 0
    items = entities if type(entities)==list else [entities]
    for entity in items:
      if not entity.get('name'):
        continue
      count+=1
      full_entity = self.__get_complete_entity(entity)
      self.__driver.add_cookie(full_entity)
    return count

  def replace(self,entities:list[dict]):
    '''
    Replace the current cookies with the new cookies
    Parameter:
      entities {list<dict>} : the new cookie entities - standard cookie dict feched by dirver.get_cookies()
    '''

    self.__driver.delete_all_cookies()
    for entity in entities:
      if 'expiry' in entity:
        entity['expiry'] = int(entity['expiry'])

      self.__driver.add_cookie(entity)


  def set(self,cookies):
    '''
    Add cookies to the current browsing context
    Parameter:
      cookies : a string or dict cookies' key and value 
       - 'BIDUPSID=77884ECAEE62BDD2A4A723BEF544DCB2; PSTM=1715957318'
       - {'BIDUPSID':'77884ECAEE62BDD2A4A723BEF544DCB2', 'PSTM':'1715957318'}
    Returns:
      {int} : The number of successful additions
    '''
    if not cookies:
      return None

    if type(cookies) == str:
      entities = self.__get_entity_form_string(cookies)
      return self.add(entities)
    elif type(cookies) == dict:
      entities = self.__get_entity_form_dict(cookies)
      return self.add(entities)
    else:
      return 0

  def __get_complete_entity(self,entity):
    '''
    Get the cookie complete parameter dict
    Parameter:
      entity {dict} : the cookie dict
    Returns:
      {dict} : the complete cookie entity
    '''
    default_entity = self.__get_default_entity()
    # copy the input dict
    input_entity = dict(entity)
    # make sure the value is str type
    input_entity['value'] = str(input_entity['value'])

    default_entity.update(input_entity)    
    return default_entity

  def __get_entity_form_dict(self,cookies_dict):
    '''
    Convert the cookies string to the entity list
    Parameter:
      cookies_dict {dict} : A dict containing multiple cookie keys and values, such as:
        - {'BIDUPSID':'77884ECAEE62BDD2A4A723BEF544DCB2', 'PSTM':'1715957318'}
    Returns:
      {list<dict>} : the cookie entity list
    '''
    entities = []
    for name,value in cookies_dict.items():
      entities.append({
        'name':name,
        'value':value,
      })

    return entities

  def __get_entity_form_string(self,cookies_string):
    '''
    Convert the cookies string to the entity list
    Parameter:
      cookies_string {str} : A string containing multiple cookie keys and values, such as:
        - 'BIDUPSID=77884ECAEE62BDD2A4A723BEF544DCB2; PSTM=1715957318'
    Returns:
      {list<dict>} : the cookie entity list
    '''

    entities = []
    kv_list = cookies_string.split(';')
    for k_v in kv_list:
      if not k_v:
        continue
      kv=k_v.split('=')
      # must be key=value string
      if len(kv)<2:
        continue
      name = kv[0].strip()
      value = kv[1].strip()
      entities.append({
        'name':name,
        'value':value,
      })

    return entities
  
  def remove(self,name=''):
    '''
    @description 删除cookie
    @param {string} name cookie name
    '''
    if name:
      return self.__driver.delete_cookie(name)
    else:
      return self.clear()
  
  def clear(self):
    return self.__driver.delete_all_cookies()



import sys,os,re

from blues_lib.dao.sql.TableMutator import TableMutator

class LoginMutator(TableMutator):
  
  _TABLE = 'naps_loginer'

  def __init__(self) -> None:
    super().__init__(self._TABLE)


